import itertools

import torch

from ..types._defs import Trajectory


def check_trajectories(trajectories: list[Trajectory], c: torch.Tensor | None):
    """Check if trajectories are not empty, well sorted and compatible with censoring.

    Args:
        trajectories (list[Trajectory]): The trajectories.
        c (torch.Tensor | None): The censoring times.

    Raises:
        ValueError: If some trajectory is empty.
        ValueError: If some trajectory is not sorted.
        ValueError: If some trajectory is not compatible with the censoring times.
    """
    if any(len(trajectory) == 0 for trajectory in trajectories):
        raise ValueError("Trajectories must not be empty")
    if any(
        not all(t0 <= t1 for t0, t1 in itertools.pairwise(t for t, _ in trajectory))
        for trajectory in trajectories
    ):
        raise ValueError(
            "Trajectories must be sorted by time, in ascending order. Also ensure "
            "there are no NaN values as this will trigger the check"
        )
    if c is not None and any(
        trajectory[-1][0] > c for trajectory, c in zip(trajectories, c, strict=True)
    ):
        raise ValueError("Transitions times may not be greater than censoring times")


def check_matrix_dim(flat: torch.Tensor, dim: int, precision_type: str):
    """Checks dimensions for matrix according to precision type.

    Args:
        flat (torch.Tensor): The flat tensor.
        dim (int): The dimension of the matrix.
        precision_type (str): The precision type.

    Raises:
        ValueError: If the number of elements is incompatible with precision type
            `'full'`.
        ValueError: If the number of elements is incompatible with precision type
            `'diag'`.
        ValueError: If the number of elements is not one and the precision type is
            `'spherical'`.
        ValueError: If the precision type is not valid.
    """
    match precision_type:
        case "full":
            if flat.numel() != (dim * (dim + 1)) // 2:
                raise ValueError(
                    f"{flat.numel()} is incompatible with full matrix of dimension "
                    f"{dim}"
                )
        case "diag":
            if flat.numel() != dim:
                raise ValueError(
                    f"{flat.numel()} is incompatible with diagonal matrix of dimension "
                    f"{dim}"
                )
        case "spherical":
            if flat.numel() != 1:
                raise ValueError(
                    f"Expected 1 element for flat, got {flat.numel()} for scalar "
                    f"marix of dimension {dim}"
                )
        case _:
            raise ValueError(
                f"Precision type must be be either 'full', 'diag' or 'spherical', got "
                f"{precision_type}"
            )



from collections import defaultdict
from collections.abc import Callable
from typing import Any, Self

import torch

from ..types._data import ModelDataUnchecked
from ..types._parameters import ModelParameters


class MCMCMixin:
    """Mixin for MCMC sampling."""

    params: ModelParameters
    n_chains: int
    init_step_size: float
    adapt_rate: float
    target_accept_rate: float

    def _logpdfs_fn(
        self,
        data: ModelDataUnchecked,
        b: torch.Tensor,
    ) -> torch.Tensor: ...

    def __init__(
        self,
        n_chains: int,
        init_step_size: float,
        adapt_rate: float,
        target_accept_rate: float,
        *args: Any,
        **kwargs: Any,
    ):
        """Initializes the MCMC mixin.

        Args:
            n_chains (int): The number of parallel chains to spawn.
            init_step_size (float): Kernel step in Metropolis-Hastings.
            adapt_rate (float): Adaptation rate for the step_size.
            target_accept_rate (float): Mean acceptance target.
        """
        super().__init__(*args, **kwargs)

        self.n_chains = n_chains
        self.init_step_size = init_step_size
        self.adapt_rate = adapt_rate
        self.target_accept_rate = target_accept_rate

    def _init_sampler(self, data: ModelDataUnchecked):
        """Initializes the MCMC sampler.

        Args:
            data (ModelDataUnchecked): The model data.

        Returns:
            MetropolisWithinGibbsSampler: The initialized MCMC sampler.
        """
        return MetropolisWithinGibbsSampler(
            lambda b: self._logpdfs_fn(data, b),
            torch.zeros(self.n_chains, len(data), self.params.random_prec.dim),
            self.n_chains,
            self.init_step_size,
            self.adapt_rate,
            self.target_accept_rate,
        )


class MetropolisWithinGibbsSampler:
    """A robust Metropolis-within-Gibbs sampler with adaptive step size."""

    logpdfs_fn: Callable[[torch.Tensor], torch.Tensor]
    n_chains: int
    adapt_rate: float
    target_accept_rate: float
    b: torch.Tensor
    logpdfs: torch.Tensor
    j: int
    step_sizes: torch.Tensor
    diagnostics_: defaultdict[str, list[torch.Tensor]]
    _noise: torch.Tensor

    def __init__(
        self,
        logpdfs_fn: Callable[[torch.Tensor], torch.Tensor],
        init_b: torch.Tensor,
        n_chains: int,
        init_step_size: float,
        adapt_rate: float,
        target_accept_rate: float,
    ):
        """Initializes the Metropolis-within-Gibbs sampler kernel.

        Args:
            logpdfs_fn (Callable[[torch.Tensor], torch.Tensor]): The log pdfs function.
            init_b (torch.Tensor): Starting b for the chain.
            n_chains (int): The number of parallel chains to spawn.
            init_step_size (float): Kernel step in Metropolis-within-Gibbs.
            adapt_rate (float): Adaptation rate for the step_size.
            target_accept_rate (float): Mean acceptance target.
        """
        self.logpdfs_fn = logpdfs_fn
        self.n_chains = n_chains
        self.adapt_rate = adapt_rate
        self.target_accept_rate = target_accept_rate

        # Set state, log pdfs, and individual parameters
        self.b = init_b
        self.reset()

        # Initialize diagnostics
        self.diagnostics_ = defaultdict(list)

        # Initialize step sizes and noise
        self.j = 0
        self.step_sizes = torch.full((1, *self.b.shape[1:]), init_step_size)
        self._noise = torch.empty(*self.b.shape[:-1])

    @torch.no_grad()  # type: ignore
    def reset(self) -> Self:
        """Resets the log pdfs and individual parameters.

        Returns:
            Self: The sampler instance.
        """
        self.logpdfs = self.logpdfs_fn(self.b)
        return self

    @torch.no_grad()  # type: ignore
    def step(self) -> Self:
        """Performs a single kernel step.

        Returns:
            Self: The sampler instance.
        """
        # Generate proposal noise
        self._noise.uniform_(-1, 1)

        # Get the proposal
        proposed_state = self.b.clone()
        proposed_state[..., self.j] += self._noise * self.step_sizes[..., self.j]
        proposed_logpdfs = self.logpdfs_fn(proposed_state)
        logpdf_diff = proposed_logpdfs - self.logpdfs

        # Vectorized acceptance decision
        log_uniform = torch.log(torch.rand_like(logpdf_diff))
        accept_mask = log_uniform < logpdf_diff

        torch.where(accept_mask.unsqueeze(-1), proposed_state, self.b, out=self.b)
        torch.where(accept_mask, proposed_logpdfs, self.logpdfs, out=self.logpdfs)

        # Update step sizes
        mean_accept_mask = accept_mask.to(torch.get_default_dtype()).mean(dim=0)
        adaptation = (mean_accept_mask - self.target_accept_rate) * self.adapt_rate
        self.step_sizes[..., self.j] *= torch.exp(adaptation)
        self.j = (self.j + 1) % self.b.size(-1)

        # Update diagnostics
        self.diagnostics_["mean_accept_rate"].append(mean_accept_mask.mean())
        self.diagnostics_["mean_step_size"].append(self.step_sizes.mean(dim=(0, 1)))

        return self

    @torch.no_grad()  # type: ignore
    def run(self, n_steps: int) -> Self:
        """Runs the sampler for a given number of steps.

        Args:
            n_steps (int): The number of steps to run the sampler for.

        Returns:
            Self: The sampler instance.
        """
        for _ in range(n_steps):
            self.step()
        return self



from typing import Any

import torch

from ..types._parameters import ModelParameters


class PriorMixin:
    """Mixin class for prior model computations."""

    params: ModelParameters

    def __init__(self, *args: Any, **kwargs: Any):
        """Initializes the prior mixin."""
        super().__init__(*args, **kwargs)

    def _prior_logliks(self, b: torch.Tensor) -> torch.Tensor:
        """Computes the prior log likelihoods.

        Args:
            b (torch.Tensor): The 3D tensor of random effects.

        Returns:
            torch.Tensor: The computed log likelihoods.
        """
        random_prec_cholesky, random_prec_log_eigvals = (
            self.params.random_prec._prec_cholesky_and_log_eigvals  # type: ignore
        )
        random_quad_form = (b @ random_prec_cholesky).pow(2).sum(dim=-1)
        random_norm_factor = random_prec_log_eigvals.sum()

        return 0.5 * (random_norm_factor - random_quad_form)



"""Utility functions for the jmstate package."""

from ._plot import plot_mcmc_diagnostics, plot_params_history
from ._surv import build_buckets

__all__ = [
    "build_buckets",
    "plot_mcmc_diagnostics",
    "plot_params_history",
]



from __future__ import annotations

import math
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
import torch
from numpy import atleast_1d
from sklearn.utils._param_validation import validate_params  # type: ignore

if TYPE_CHECKING:
    from ..model._base import MultiStateJointModel


@validate_params(
    {
        "figsize": [tuple],
    },
    prefer_skip_nested_validation=True,
)
def plot_params_history(
    model: MultiStateJointModel,
    *,
    figsize: tuple[int, int] = (10, 8),
) -> tuple[plt.Figure, np.ndarray]:  # type: ignore
    r"""Visualize the evolution of model parameters during fitting.

    This function generates a grid of subplots showing the trajectory of each model
    parameter across iterations, allowing assessment of convergence and exploration
    of the optimization process.

    Args:
        model (MultiStateJointModel): The fitted model whose parameter history is to
            be plotted.
        figsize (tuple[int, int], optional): Figure dimensions `(width, height)`.
            Defaults to `(10, 8)`.

    Raises:
        ValueError: If the model contains fewer than two recorded parameter states,
            preventing visualization.

    Returns:
        tuple[plt.Figure, np.ndarray]: A tuple containing the matplotlib `Figure`
            object and a flattened array of `Axes` objects corresponding to the
            subplots.
    """
    from ..model._base import MultiStateJointModel  # noqa: PLC0415

    validate_params(
        {
            "model": [MultiStateJointModel],
        },
        prefer_skip_nested_validation=True,
    )

    if len(model.params_history_) <= 1:
        raise ValueError("More than one recorded parameter is required to plot")

    # Get the names
    named_parameters_dict = dict(model.params.named_parameters())
    nsubplots = len(named_parameters_dict)
    ncols = math.ceil(math.sqrt(nsubplots))
    nrows = math.ceil(nsubplots / ncols)

    # Create subplots
    fig, axes = plt.subplots(nrows, ncols, figsize=figsize)  # type: ignore
    axes = atleast_1d(axes).ravel()

    Y = torch.stack(model.params_history_)
    i = 0
    for ax, (name, val) in zip(axes, named_parameters_dict.items(), strict=False):
        history = Y[:, i : (i := i + val.numel())]
        ax.plot(history, label=[f"{name}[{j}]" for j in range(val.numel())])
        ax.set(title=name, xlabel="Iteration", ylabel="Value")
        ax.legend()

    # Remove unused subplots
    for j in range(nsubplots, len(axes)):
        fig.delaxes(axes[j])

    plt.suptitle("Stochastic optimization of the parameters")  # type: ignore
    plt.tight_layout()

    return fig, axes


@validate_params(
    {
        "figsize": [tuple],
    },
    prefer_skip_nested_validation=True,
)
def plot_mcmc_diagnostics(
    model: MultiStateJointModel,
    *,
    figsize: tuple[int, int] = (8, 4),
) -> tuple[plt.Figure, np.ndarray]:  # type: ignore
    r"""Visualize the evolution of MCMC diagnostics during fitting.

    This function generates two subplots showing the mean acceptance rate and mean step
    size across iterations.

    Args:
        model (MultiStateJointModel): The fitted model whose MCMC diagnostics are to be
            plotted.
        figsize (tuple[int, int], optional): Figure dimensions `(width, height)`.
            Defaults to `(8, 4)`.

    Raises:
        ValueError: If the model sampler is `None` or the diagnostics have fewer than
            two recorded values.

    Returns:
        tuple[plt.Figure, np.ndarray]: A tuple containing the matplotlib `Figure`
            object and a flattened array of `Axes` objects corresponding to the
            subplots.
    """
    from ..model._base import MultiStateJointModel  # noqa: PLC0415

    validate_params(
        {
            "model": [MultiStateJointModel],
        },
        prefer_skip_nested_validation=True,
    )

    if model.sampler is None:
        raise ValueError("Model sampler is None")

    if len(model.sampler.diagnostics_["mean_accept_rate"]) <= 1:
        raise ValueError("More than one recorded MCMC step is required to plot")

    # Create subplots
    fig, axes = plt.subplots(1, 2, figsize=figsize)  # type: ignore
    axes = atleast_1d(axes).ravel()

    axes[0].plot(model.sampler.diagnostics_["mean_accept_rate"])
    axes[0].set(
        title="Mean acceptance rate",
        xlabel="Iteration",
        ylabel="Mean acceptance rate",
    )
    axes[0].axhline(
        y=model.sampler.target_accept_rate,
        color="r",
        linestyle="--",
        label="Target acceptance rate",
    )

    axes[1].plot(
        model.sampler.diagnostics_["mean_step_size"],
        label=[f"b[{j}]" for j in range(model.sampler.b.size(-1))],
    )
    axes[1].set(
        title="Mean step size (component-wise)",
        xlabel="Iteration",
        ylabel="Mean step size",
    )

    for ax in axes:
        ax.axvline(x=model.n_warmup, color="gray", linestyle="--", label="Warmup end")
        ax.legend()

    plt.suptitle("MCMC diagnostics")  # type: ignore
    plt.tight_layout()

    return fig, axes



"""Base hazard functions."""

__all__ = ["Exponential", "Gompertz", "LogNormal", "Weibull"]

from numbers import Real
from typing import cast

import torch
from sklearn.utils._param_validation import (  # type: ignore
    Interval,
    StrOptions,
    validate_params,  # type: ignore
)
from torch import nn

from ..types._defs import LOG_TWO_PI, LogBaseHazardFn


class Exponential(LogBaseHazardFn):
    r"""Implements the Exponential base hazard.

    Exponential base hazard is time independent.

    It is given by the formula:

    .. math::
        \lambda_0(t) = \lambda.

    This method expects:
        - `t0`: a column vector of previous transition times, shape `(n, 1)`.
        - `t1`: a matrix of future evaluation times, shape `(n, m)`, with the same
          number of rows as `t0`.

    The output is the log base hazard evaluated at each `t1` relative to `t0`.

    Optimization of the parameters can be disabled by checking the `forzen` flag.

    Attributes:
        log_lmda (nn.Parameter | torch.Tensor): The log rate factor.
        frozen (bool): Whether the parameters are frozen.
    """

    log_lmda: nn.Parameter | torch.Tensor

    @validate_params(
        {
            "lmda": [Interval(Real, 0, None, closed="neither")],
            "frozen": [bool],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(self, lmda: float, *, frozen: bool = False):
        """Initializes the Exponential hazard.

        Args:
            lmda (float): The rate factor.
            frozen (bool, optional): Whether to freeze the parameters. Defaults to
                `False`.
        """
        super().__init__()  # type: ignore

        log_lmda_tensor = torch.log(torch.tensor(lmda))
        self.log_lmda = nn.Parameter(log_lmda_tensor) if not frozen else log_lmda_tensor
        self.frozen = frozen

    def forward(
        self,
        t0: torch.Tensor,  # noqa: ARG002
        t1: torch.Tensor,  # noqa: ARG002
    ) -> torch.Tensor:
        """Calls the Exponential base hazard.

        Args:
            t0 (torch.Tensor): Previous transition times, shape :math:`(n, 1)`.
            t1 (torch.Tensor): Future evaluation times, shape :math:`(n, m)`.

        Returns:
            torch.Tensor: The computed base hazard in log scale.
        """
        return self.log_lmda

    @property
    def lmda(self) -> torch.Tensor:
        """Gets the rate factor.

        Returns:
            torch.Tensor: The rate factor.
        """
        return self.log_lmda.exp()


class Weibull(LogBaseHazardFn):
    r"""Implements the Weibull base hazard.

    Weibull base hazard is time dependent.

    It is given by the formula:

    .. math::
        \lambda_0(t) = k \lambda^k t^{k - 1}.

    This method expects:
        - `t0`: a column vector of previous transition times, shape `(n, 1)`.
        - `t1`: a matrix of future evaluation times, shape `(n, m)`, with the same
            number of rows as `t0`.

    The output is the log base hazard evaluated at each `t1` relative to `t0`.

    If `clock_type` is set to `sojourn`, given `t0` and `t1`, the transformation will be
    computed at `t1 - t0` (sojourn time), and simply `t1` if set to `absolute`.

    Optimization of the parameters can be disabled by checking the `forzen` flag.

    Attributes:
        log_lmda (nn.Parameter | torch.Tensor): The log of the scale parameter.
        log_k (nn.Parameter | torch.Tensor): The log of the shape parameter.
        clock_type (str): The type of clock to use.
        frozen (bool): Whether the parameters are frozen.
    """

    log_lmda: nn.Parameter | torch.Tensor
    log_k: nn.Parameter | torch.Tensor
    clock_type: str
    frozen: bool

    @validate_params(
        {
            "lmda": [Interval(Real, 0, None, closed="neither")],
            "k": [Interval(Real, 0, None, closed="neither")],
            "clock_type": [StrOptions({"sojourn", "absolute"})],
            "frozen": [bool],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(
        self,
        lmda: float,
        k: float,
        *,
        clock_type: str = "sojourn",
        frozen: bool = False,
    ):
        """Initializes the Weibull base hazard.

        Args:
            lmda (float): The scale parameter.
            k (float): The shape parameter.
            clock_type (str, optional): The type of clock to use. Defaults to
                "sojourn".
            frozen (bool, optional): Whether to freeze the parameters. Defaults to
                `False`.
        """
        super().__init__()  # type: ignore

        log_lmda_tensor = torch.log(torch.tensor(lmda))
        self.log_lmda = nn.Parameter(log_lmda_tensor) if not frozen else log_lmda_tensor
        log_k_tensor = torch.log(torch.tensor(k))
        self.log_k = nn.Parameter(log_k_tensor) if not frozen else log_k_tensor
        self.clock_type = clock_type
        self.frozen = frozen

    def forward(self, t0: torch.Tensor, t1: torch.Tensor) -> torch.Tensor:
        """Calls the Weibull base hazard.

        Args:
            t0 (torch.Tensor): Previous transition times, shape `(n, 1)`.
            t1 (torch.Tensor): Future evaluation times, shape `(n, m)`.

        Returns:
            torch.Tensor: The computed base hazard in log scale.
        """
        t = t1 - t0 if self.clock_type == "sojourn" else t1
        log_t = torch.log(t).clamp(min=-50)
        return self.log_k + self.k * self.log_lmda + (self.k - 1) * log_t

    @property
    def k(self) -> torch.Tensor:
        """Gets the shape parameter.

        Returns:
            torch.Tensor: The shape parameter.
        """
        return self.log_k.exp()

    @property
    def lmda(self) -> torch.Tensor:
        """Gets the scale parameter.

        Returns:
            torch.Tensor: The scale parameter.
        """
        return self.log_lmda.exp()


class Gompertz(LogBaseHazardFn):
    r"""Implements the Gompertz base hazard.

    Gompertz base hazard is time dependent.
    It is given by the formula:

    .. math::
        \lambda_0(t) = a \exp{bt}.

    This method expects:
        - `t0`: a column vector of previous transition times, shape `(n, 1)`.
        - `t1`: a matrix of future evaluation times, shape `(n, m)`, with the same
          number of rows as `t0`.

    The output is the log base hazard evaluated at each `t1` relative to `t0`.

    If `clock_type` is set to `sojourn`, given `t0` and `t1`, the transformation will be
    computed at `t1 - t0` (sojourn time), and simply `t1` if `clock_type` is set to
    `absolute`.

    Optimization of the parameters can be disabled by checking the `forzen` flag.

    Attributes:
        log_a (nn.Parameter | torch.Tensor): The baseline hazard parameter.
        b (nn.Parameter | torch.Tensor): The shape parameter.
        clock_type (str): The type of clock to use.
        frozen (bool): Whether the parameters are frozen.
    """

    log_a: nn.Parameter | torch.Tensor
    b: nn.Parameter | torch.Tensor
    clock_type: str
    frozen: bool

    @validate_params(
        {
            "a": [Interval(Real, 0, None, closed="neither")],
            "b": [Interval(Real, None, None, closed="neither")],
            "clock_type": [StrOptions({"sojourn", "absolute"})],
            "frozen": [bool],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(
        self,
        a: float,
        b: float,
        *,
        clock_type: str = "sojourn",
        frozen: bool = False,
    ):
        """Initializes the Gompertz base hazard.

        Args:
            a (float): The baseline hazard.
            b (float): The shape parameter.
            clock_type (str, optional): The type of clock to use. Defaults to
                "sojourn".
            frozen (bool, optional): Whether to freeze the parameters. Defaults to
                `False`.
        """
        super().__init__()  # type: ignore

        log_a_tensor = torch.log(torch.tensor(a))
        self.log_a = nn.Parameter(log_a_tensor) if not frozen else log_a_tensor
        b_tensor = torch.tensor(b)
        self.b = nn.Parameter(b_tensor) if not frozen else b_tensor
        self.clock_type = clock_type
        self.frozen = frozen

    def forward(self, t0: torch.Tensor, t1: torch.Tensor) -> torch.Tensor:
        """Calls the Gompertz base hazard.

        Args:
            t0 (torch.Tensor): Previous transition times, shape `(n, 1)`.
            t1 (torch.Tensor): Future evaluation times, shape `(n, m)`.

        Returns:
            torch.Tensor: The computed base hazard in log scale.
        """
        t = t1 - t0 if self.clock_type == "sojourn" else t1
        return self.log_a + self.b * t

    @property
    def a(self) -> torch.Tensor:
        """Gets the baseline hazard.

        Returns:
            torch.Tensor: The baseline hazard.
        """
        return self.log_a.exp()


class LogNormal(LogBaseHazardFn):
    r"""Implements the log normal base hazard.

    Log normal base hazard is time dependent.
    It is given by the formula:

    .. math::
        \lambda_0(t) = \frac{\phi\left( \frac{\log t - \mu}{\sigma} \right)}{t \sigma
        \, \Phi\left( -\frac{\log t - \mu}{\sigma} \right)},
        \quad t > 0,

    where:

    .. math::
        \phi(z) = \frac{1}{\sqrt{2\pi}} e^{-z^2/2}, \quad
        \Phi(z) = \frac{1}{\sqrt{2\pi}} \int_{-\infty}^z e^{-t^2/2} \, dt.

    This method expects:
        - `t0`: a column vector of previous transition times, shape `(n, 1)`.
        - `t1`: a matrix of future evaluation times, shape `(n, m)`, with the same
          number of rows as `t0`.

    The output is the log base hazard evaluated at each `t1` relative to `t0`.

    If `clock_type` is set to `sojourn`, given `t0` and `t1`, the transformation will be
    computed at `t1 - t0` (sojourn time), and simply `t1` if `clock_type` is set to
    `absolute`.

    Optimization of the parameters can be disabled by checking the `forzen` flag.

    Attributes:
        mu (nn.Parameter | torch.Tensor): The log time mean.
        log_scale (nn.Parameter | torch.Tensor): The log of scale.
        clock_type (str): The type of clock to use.
        frozen (bool): Whether the parameters are frozen.
    """

    mu: nn.Parameter | torch.Tensor
    log_scale: nn.Parameter | torch.Tensor
    clock_type: str
    frozen: bool

    @validate_params(
        {
            "mu": [Interval(Real, None, None, closed="neither")],
            "scale": [Interval(Real, 0, None, closed="neither")],
            "clock_type": [StrOptions({"sojourn", "absolute"})],
            "frozen": [bool],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(
        self,
        mu: float,
        scale: float,
        *,
        clock_type: str = "sojourn",
        frozen: bool = False,
    ):
        """Initializes the log normal base hazard.

        Args:
            mu (float): The log time mean.
            scale (float): The log time scale.
            clock_type (str, optional): The type of clock to use. Defaults to
                "sojourn".
            frozen (bool, optional): Whether to freeze the parameters. Defaults to
                `False`.
        """
        super().__init__()  # type: ignore

        mu_tensor = torch.tensor(mu)
        self.mu = nn.Parameter(mu_tensor) if not frozen else mu_tensor
        log_scale_tensor = torch.log(torch.tensor(scale))
        self.log_scale = (
            nn.Parameter(log_scale_tensor) if not frozen else log_scale_tensor
        )
        self.clock_type = clock_type
        self.frozen = frozen

    def forward(self, t0: torch.Tensor, t1: torch.Tensor) -> torch.Tensor:
        """Calls the log normal base hazard.

        Args:
            t0 (torch.Tensor): Previous transition times, shape `(n, 1)`.
            t1 (torch.Tensor): Future evaluation times, shape `(n, m)`.

        Returns:
            torch.Tensor: The computed base hazard in log scale.
        """
        t = t1 - t0 if self.clock_type == "sojourn" else t1
        log_t = torch.log(t).clamp(min=-50)
        z = (log_t - self.mu) / self.scale
        log_pdf = -log_t - self.log_scale - 0.5 * LOG_TWO_PI - 0.5 * z**2
        log_sf = cast(torch.Tensor, torch.special.log_ndtr(-z))  # type: ignore
        return log_pdf - log_sf

    @property
    def scale(self) -> torch.Tensor:
        """Gets the scale.

        Returns:
            torch.Tensor: The scale.
        """
        return self.log_scale.exp()



from bisect import bisect_left
from numbers import Integral, Real
from typing import Final

import torch
from rich.console import Console, Group
from rich.panel import Panel
from rich.rule import Rule
from rich.table import Table
from rich.text import Text
from sklearn.base import BaseEstimator  # type: ignore
from sklearn.utils._param_validation import Interval, validate_params  # type: ignore
from torch.distributions import Normal
from torch.nn.utils import parameters_to_vector

from ..types._data import ModelDesign
from ..types._parameters import ModelParameters
from ._fit import FitMixin
from ._predict import PredictMixin
from ._sampler import MetropolisWithinGibbsSampler

# Constants
SIGNIFICANCE_LEVELS: Final[tuple[float, ...]] = (
    0.001,
    0.01,
    0.05,
    0.1,
    float("inf"),
)
SIGNIFICANCE_CODES: Final[tuple[str, ...]] = (
    "[red3]***[/]",
    "[orange3]**[/]",
    "[yellow3]*[/]",
    ".",
    "",
)


class MultiStateJointModel(BaseEstimator, FitMixin, PredictMixin):
    r"""Nonlinear multistate joint model for longitudinal and survival data.

    This class generalizes both linear and standard joint models to accommodate
    multiple states under a semi-Markov assumption. The model is fully specified
    by a `ModelDesign` object, which defines longitudinal and hazard functions, and
    a `ModelParameters` object, which contains the associated (modifiable) parameters.
    Parameters may be shared across functions as specified in the model design. The
    `ModelDesign` is fixed at initialization, while `ModelParameters` are updated
    in place during fitting.

    Model fitting is performed using a stochastic gradient ascent algorithm, and
    parameter sampling is handled by a Metropolis-Within-Gibbs MCMC procedure. The
    default settings are based on commonly accepted values, and the step sizes are
    adapted component-wise dynamically based on the acceptance rate.

    Dynamic prediction is supported via the prediction methods in `PredictMixin`,
    which allow both single and double Monte Carlo integration.

    Design settings:
        - `design`: The model specification defining the individual parameters,
          regression, and link functions.
        - `params`: Initial values for the model parameters; modifiable during fitting.

    Numerical integration settings:
        - `n_quad`: Number of nodes for Gauss-Legendre quadrature in hazard integration.
        - `n_bisect`: Number of bisection steps for transition time sampling.

    MCMC settings:
        - `sampler`: MCMC sampler instance used during model fitting.
        - `n_chains`: Number of parallel MCMC chains.
        - `init_step_size`: Initial kernel standard deviation in
           Metropolis-Within-Gibbs.
        - `adapt_rate`: Adaptation rate for the step size.
        - `target_accept_rate`: Target mean acceptance probability.
        - `n_warmup`: Number of warmup iterations per chain.
        - `n_subsample`: Number of subsamples between predictions; higher values
          reduce autocorrelation but increase computation time. A value of one means
          no subsampling. This value may be very sensitive.

    Fitting settings:
        - `optimizer`: Optimizer for stochastic gradient ascent. If `None`, fitting
          is disabled. Recommended: `torch.optim.Adam` with learning rate 0.01 to 0.5.
        - `max_iter`: Maximum iterations for gradient ascent.
        - `tol`: Tolerance for the :math:`R^2` convergence criterion.
        - `window_size`: Window size for :math:`R^2` convergence; default 100.
          This criterion is scale-agnostic and provides a local stationarity test.

    Printing and visualization:
        - `verbose`: Whether to print progress during fitting and prediction.
        - After fitting, `summary` and `plot_params_history` (from `jmstate.utils`)
          can be used to display p-values, log-likelihood, AIC, BIC, and the
          evolution of parameters over iterations.

    Attributes:
        design (ModelDesign): The model specification defining the individual
            parameters, regression, and link functions.
        params (ModelParameters): The modifiable model parameters.
        optimizer (torch.optim.Optimizer | None): Optimizer used for fitting.
        n_quad (int): Number of Gauss-Legendre quadrature nodes for hazard integration.
        n_bisect (int): Number of bisection steps for transition time sampling.
        sampler (MetropolisWithinGibbsSampler | None): MCMC sampler instance used during
            model fitting.
        n_chains (int): Number of parallel MCMC chains.
        init_step_size (float): Initial kernel standard deviation in
            Metropolis-Within-Gibbs.
        adapt_rate (float): Adaptation rate for the MCMC step size.
        target_accept_rate (float): Target acceptance probability.
        n_warmup (int): Number of warmup iterations per MCMC chain.
        n_subsample (int): Number of subsamples for MCMC iterations.
        max_iter (int): Maximum number of iterations for stochastic gradient ascent.
        tol (float): Tolerance for :math:`R^2` convergence criterion.
        window_size (int): Window size for :math:`R^2` convergence evaluation.
        verbose (bool): Flag to print fitting and prediction progress.
        params_history_ (list[torch.Tensor]): History of parameter values as flattened
            tensors.
        fim_ (torch.Tensor | None): Fisher Information Matrix.
        loglik_ (float | None): Log-likelihood of the fitted model.
        aic_ (float | None): Akaike Information Criterion.
        bic_ (float | None): Bayesian Information Criterion.

    Examples:
        >>> from jmstate import MultiStateJointModel
        >>> optimizer = torch.optim.Adam(params.parameters(), lr=0.1)
        >>> model = MultiStateJointModel(design, params, optimizer)
        >>> model.fit(data)
        >>> model.compute_summary().summary()
    """

    design: ModelDesign
    params: ModelParameters
    optimizer: torch.optim.Optimizer | None
    n_quad: int
    n_bisect: int
    sampler: MetropolisWithinGibbsSampler | None
    n_chains: int
    init_step_size: float
    adapt_rate: float
    target_accept_rate: float
    n_warmup: int
    n_subsample: int
    max_iter: int
    tol: float
    window_size: int
    verbose: bool | int
    params_history_: list[torch.Tensor]
    fim_: torch.Tensor | None
    loglik_: float | None
    aic_: float | None
    bic_: float | None

    @validate_params(
        {
            "design": [ModelDesign],
            "params": [ModelParameters],
            "optimizer": [torch.optim.Optimizer, None],
            "n_quad": [Interval(Integral, 1, None, closed="left")],
            "n_bisect": [Interval(Integral, 1, None, closed="left")],
            "n_chains": [Interval(Integral, 1, None, closed="left")],
            "init_step_size": [Interval(Real, 0, None, closed="neither")],
            "adapt_rate": [Interval(Real, 0, None, closed="left")],
            "target_accept_rate": [Interval(Real, 0, 1, closed="neither")],
            "n_warmup": [Interval(Integral, 0, None, closed="left")],
            "n_subsample": [Interval(Integral, 0, None, closed="left")],
            "max_iter": [Interval(Integral, 0, None, closed="left")],
            "tol": [Interval(Real, 0, 1, closed="both")],
            "window_size": [Interval(Integral, 2, None, closed="left")],
            "verbose": ["verbose"],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(
        self,
        design: ModelDesign,
        params: ModelParameters,
        optimizer: torch.optim.Optimizer | None = None,
        *,
        n_quad: int = 32,
        n_bisect: int = 32,
        n_chains: int = 5,
        init_step_size: float = 0.1,
        adapt_rate: float = 0.1,
        target_accept_rate: float = 0.44,
        n_warmup: int = 100,
        n_subsample: int = 10,
        max_iter: int = 1000,
        tol: float = 0.1,
        window_size: int = 100,
        verbose: bool | int = True,
    ):
        r"""Initialize the multistate joint model with specified design and parameters.

        Constructs a joint model based on a user-defined `ModelDesign` and initial
        `ModelParameters`. Provides default settings for numerical integration, MCMC
        sampling, stochastic gradient fitting, and printing options.

        Args:
            design (ModelDesign): The model specification defining the individual
                parameters, regression, and link functions.
            params (ModelParameters): Initial values for the model parameters;
                modifiable during fitting.
            optimizer (torch.optim.Optimizer | None, optional): Optimizer used for
                fitting. If `None`, fitting is disabled. Defaults to None.
            n_quad (int, optional): Number of nodes for Gauss-Legendre quadrature in
                hazard integration. Defaults to 32.
            n_bisect (int, optional): Number of bisection steps for transition time
                sampling. Defaults to 32.
            n_chains (int, optional): Number of parallel MCMC chains. Defaults to 5.
            init_step_size (float, optional): Initial step size for the MCMC sampler.
                Defaults to 0.1.
            adapt_rate (float, optional): Adaptation rate for the MCMC step size.
                Defaults to 0.1.
            target_accept_rate (float, optional): Target mean acceptance probability for
                Metropolis-Within-Gibbs. Defaults to 0.44.
            n_warmup (int, optional): Number of warmup iterations per MCMC chain.
                Defaults to 100.
            n_subsample (int, optional): Number of subsamples between MCMC updates.
                Defaults to 10.
            max_iter (int, optional): Maximum number of iterations for stochastic
                gradient ascent. Defaults to 1000.
            tol (float, optional): Tolerance for :math:`R^2` convergence criterion.
                Defaults to 0.1.
            window_size (int, optional): Window size for :math:`R^2` convergence
                evaluation. Defaults to 100.
            verbose (bool | int, optional): Flag to print progress during fitting and
                prediction. Defaults to True.
        """
        # Info of the Mixin Classes
        super().__init__(
            optimizer=optimizer,
            n_quad=n_quad,
            n_bisect=n_bisect,
            n_chains=n_chains,
            init_step_size=init_step_size,
            adapt_rate=adapt_rate,
            target_accept_rate=target_accept_rate,
            max_iter=max_iter,
            tol=tol,
            window_size=window_size,
        )

        # Store model components
        self.design = design
        self.params = params
        self.n_warmup = n_warmup
        self.n_subsample = n_subsample
        self.verbose = verbose
        self.params_history_ = [parameters_to_vector(self.params.parameters()).detach()]
        self.fim_ = None
        self.loglik_ = None
        self.aic_ = None
        self.bic_ = None

    @property
    def stderr(self) -> torch.Tensor:
        r"""Computes the estimated standard errors of the model parameters.

        The standard errors are derived from the diagonal of the inverse of the
        estimated Fisher Information Matrix evaluated at the Maximum Likelihood Estimate
        (MLE). They provide a measure of uncertainty for each parameter and can be used
        to construct confidence intervals.

        .. math::
            \mathrm{stderr} = \sqrt{\operatorname{diag}\left( \hat{\mathcal{I}}_n
            (\hat{\theta})^{-1} \right)}

        Raises:
            ValueError: If the model has not been fitted and the Fisher Information
                Matrix is unavailable.

        Returns:
            torch.Tensor: Vector of standard errors corresponding to each parameter.
        """
        if self.fim_ is None:
            raise ValueError(
                "Fisher Information Matrix is not available. Call "
                "compute_summary() first."
            )

        return self.fim_.inverse().diag().sqrt()

    def summary(self):
        r"""Print a statistical summary of the fitted multistate joint model.

        This function displays the estimated parameter values, their standard errors,
        and the (nullity) p-values testing the hypothesis that each coefficient is zero.
        Note that for certain parameters—such as precision components—these p-values
        may not be meaningful and should be interpreted with caution.

        Additionally, the function prints model selection metrics including the log
        likelihood, Akaike Information Criterion (AIC), and Bayesian Information
        Criterion (BIC), with color-enhanced formatting for readability.

        Raises:
            ValueError: If the model has not been fitted.
        """
        vector = parameters_to_vector(self.params.parameters())
        stderr = self.stderr
        zvalues = torch.abs(vector / stderr)
        pvalues = 2 * (1 - Normal(0, 1).cdf(zvalues))

        table = Table()
        table.add_column("Parameter name", justify="left")
        table.add_column("Value", justify="center")
        table.add_column("Standard Error", justify="center")
        table.add_column("z-value", justify="center")
        table.add_column("p-value", justify="center")
        table.add_column("Significance level", justify="center")

        i = 0
        for name, val in self.params.named_parameters():
            for j in range(val.numel()):
                code = SIGNIFICANCE_CODES[
                    bisect_left(SIGNIFICANCE_LEVELS, pvalues[i].item())
                ]

                table.add_row(
                    f"{name}[{j}]",
                    f"{vector[i].item():.3f}",
                    f"{stderr[i].item():.3f}",
                    f"{zvalues[i].item():.3f}",
                    f"{pvalues[i].item():.3f}",
                    code,
                )
                i += 1

        criteria = Text(
            f"Log-likelihood: {self.loglik_:.3f}\n"
            f"AIC: {self.aic_:.3f}\n"
            f"BIC: {self.bic_:.3f}",
            style="bold cyan",
        )

        content = Group(table, Rule(style="dim"), criteria, Rule(style="dim"))
        panel = Panel(
            content, title="Model Summary", border_style="green", expand=False
        )

        Console().print(panel)



"""Type definitions for the jmstate package."""

from ._data import ModelData, ModelDesign, SampleData
from ._defs import (
    BucketData,
    IndividualParametersFn,
    LinkFn,
    LogBaseHazardFn,
    RegressionFn,
)
from ._parameters import ModelParameters, PrecisionParameters

__all__ = [
    "BucketData",
    "IndividualParametersFn",
    "LinkFn",
    "LogBaseHazardFn",
    "ModelData",
    "ModelDesign",
    "ModelParameters",
    "PrecisionParameters",
    "RegressionFn",
    "SampleData",
]



import torch


def _tril_from_flat(flat: torch.Tensor, dim: int) -> torch.Tensor:
    """Generates the lower triangular matrix associated with flat tensor.

    Args:
        flat (torch.Tensor): Flat tensor
        dim (int): Dimension of the matrix.

    Returns:
        torch.Tensor: The lower triangular matrix.
    """
    return torch.zeros(dim, dim).index_put_(tuple(torch.tril_indices(dim, dim)), flat)


def _flat_from_tril(L: torch.Tensor) -> torch.Tensor:
    """Flattens the lower triangular part (including the diagonal) of a square matrix.

    Into a 1D tensor, in row-wise order.

    Args:
        L (torch.Tensor): Square lower-triangular matrix of shape (dim, dim).

    Returns:
        torch.Tensor: Flattened 1D tensor containing the lower triangular entries.
    """
    dim = L.size(0)
    return L[tuple(torch.tril_indices(dim, dim))]


def log_cholesky_from_flat(
    flat: torch.Tensor, dim: int, precision_type: str
) -> torch.Tensor:
    """Computes log cholesky from flat tensor according to choice of precision type.

    Args:
        flat (torch.Tensor): The flat tensor parameter.
        dim (int): The dimension of the matrix.
        precision_type (str): The precision type, `'full'`, `'diag'`, or
            `'spherical'`.

    Raises:
        ValueError: If the precision type is not valid.

    Returns:
        torch.Tensor: The log cholesky representation.
    """
    match precision_type:
        case "full":
            return _tril_from_flat(flat, dim)
        case "diag":
            return torch.diag(flat)
        case "spherical":
            return flat * torch.eye(dim)
        case _:
            raise ValueError(
                "Precision type must be be either 'full', 'diag' or 'spherical', got "
                f"{precision_type}"
            )


def flat_from_log_cholesky(L: torch.Tensor, precision_type: str) -> torch.Tensor:
    """Computes flat tensor from log cholesky according to choice of precision type.

    Args:
        L (torch.Tensor): The square lower triangular matrix parameter.
        precision_type (str): The precision type, `'full'`, `'diag'`, or
            `'spherical'`.

    Raises:
        ValueError: If the precision type is not valid.

    Returns:
        torch.Tensor: The flat representation.
    """
    match precision_type:
        case "full":
            return _flat_from_tril(L)
        case "diag":
            return L.diag()
        case "spherical":
            return L[0, 0].flatten()
        case _:
            raise ValueError(
                "Precision type must be be either 'full', 'diag' or 'spherical', got "
                f"{precision_type}"
            )



from math import ceil
from numbers import Integral
from typing import Any, cast

import torch
from sklearn.utils._param_validation import Interval, validate_params  # type: ignore
from sklearn.utils.validation import (  # type: ignore
    assert_all_finite,  # type: ignore
    check_consistent_length,  # type: ignore
)
from torch.nn.utils import parameters_to_vector, vector_to_parameters
from tqdm import trange

from ..types._data import (
    ModelData,
    ModelDataUnchecked,
    ModelDesign,
    SampleData,
    SampleDataUnchecked,
)
from ..types._defs import Trajectory
from ..types._parameters import ModelParameters
from ._hazard import HazardMixin
from ._sampler import MCMCMixin


class PredictMixin(HazardMixin, MCMCMixin):
    """Mixin class for prediction."""

    design: ModelDesign
    params: ModelParameters
    n_chains: int
    n_warmup: int
    n_subsample: int
    verbose: bool | int
    fim_: torch.Tensor | None

    def __init__(self, *args: Any, **kwargs: Any):
        """Initialize the prediction mixin."""
        super().__init__(*args, **kwargs)

    def _sample_params(self, sample_size: int) -> torch.Tensor:
        """Sample model parameters based on asymptotic behavior of the MLE.

        This uses Bernstein-von Mises theorem to approximate the posterior distribution
        of the parameters as a multivariate normal distribution with mean equal to the
        MLE and covariance matrix equal to the inverse of the Fisher Information Matrix.

        Args:
            sample_size (int): The desired sample size.

        Returns:
            torch.Tensor: A tensor of sampled model parameters as vectors.
        """
        dist = torch.distributions.MultivariateNormal(
            loc=parameters_to_vector(self.params.parameters()).detach(),
            precision_matrix=cast(torch.Tensor, self.fim_),
        )
        return dist.sample((sample_size,))

    @torch.no_grad()  # type: ignore
    @validate_params(
        {
            "data": [ModelData],
            "u": [torch.Tensor],
            "n_samples": [Interval(Integral, 1, None, closed="left")],
            "double_monte_carlo": [bool],
        },
        prefer_skip_nested_validation=True,
    )
    def predict_y(
        self,
        data: ModelData,
        u: torch.Tensor,
        *,
        n_samples: int = 1000,
        double_monte_carlo: bool = False,
    ) -> torch.Tensor:
        r"""Predict longitudinal measurements at specified times.

        Computes the predicted longitudinal responses for each individual at the
        specified prediction times :math:`u` by evaluating the regression function
        conditional on posterior draws of the random effects :math:`b`. The prediction
        may optionally use a double Monte Carlo procedure to account for parameter
        uncertainty following Rizopoulos (2011).

        The input `u` must be a matrix of shape :math:`(n, m)` where :math:`n` is the
        number of individuals and :math:`m` is the number of prediction time points.

        Args:
            data (ModelData): The dataset containing covariates, observed outcomes,
                trajectories, and censoring information.
            u (torch.Tensor): Matrix of prediction times of shape `(n, m)`.
            n_samples (int, optional): Number of posterior samples to draw. Defaults to
                1000.
            double_monte_carlo (bool, optional): If True, predictions incorporate a
                double Monte Carlo procedure to sample parameters. Defaults to False.

        Raises:
            ValueError: If `double_monte_carlo` is True and the model has not been
                fitted.
            ValueError: If `u` contains NaN or infinite values.
            ValueError: If `u` has a shape inconsistent with the number of individuals.

        Returns:
            torch.Tensor: Predicted longitudinal outcomes of shape `(n_samples, n, m)`,
                where predictions are stacked along the first dimension.
        """
        assert_all_finite(u, input_name="u")
        check_consistent_length(u, data)

        # Load and complete data
        data = ModelDataUnchecked(
            data.x, data.t, data.y, data.trajectories, data.c
        ).prepare(self)

        # Initialize variables
        y_pred: list[torch.Tensor] = []
        n_iter = ceil(n_samples / self.n_chains)

        if double_monte_carlo:
            init_params = parameters_to_vector(self.params.parameters())
            sampled_params = self._sample_params(n_iter)

        # Initialize MCMC
        sampler = self._init_sampler(data).run(self.n_warmup)

        for i in trange(
            n_iter,
            desc="Predicting longitudinal values",
            disable=not bool(self.verbose),
        ):
            if double_monte_carlo:
                vector_to_parameters(sampled_params[i], self.params.parameters())  # type: ignore

            indiv_params = self.design.indiv_params_fn(
                self.params.fixed_effects, data.x, sampler.b
            )
            y = self.design.regression_fn(u, indiv_params)
            y_pred.extend(y[i] for i in range(y.size(0)))

            sampler.run(self.n_subsample)

        # Restore parameters
        if double_monte_carlo:
            vector_to_parameters(init_params, self.params.parameters())  # type: ignore

        return torch.stack(y_pred[:n_samples])

    @torch.no_grad()  # type: ignore
    @validate_params(
        {
            "data": [ModelData],
            "u": [torch.Tensor],
            "n_samples": [Interval(Integral, 1, None, closed="left")],
            "double_monte_carlo": [bool],
        },
        prefer_skip_nested_validation=True,
    )
    def predict_surv_logps(
        self,
        data: ModelData,
        u: torch.Tensor,
        *,
        n_samples: int = 1000,
        double_monte_carlo: bool = False,
    ) -> torch.Tensor:
        r"""Predict survival log-probabilities at specified times.

        Computes the predicted log survival probabilities for each individual at the
        specified prediction times :math:`u` by evaluating the survival function
        conditional on posterior draws of the random effects :math:`b`. The computation
        may optionally use a double Monte Carlo procedure to incorporate parameter
        uncertainty following Rizopoulos (2011).

        The predicted quantity is given by:

        .. math::
            \log \mathbb{P}(T^* \geq u \mid T^* > c) = -\int_c^u \lambda(t) \, dt,

        where :math:`c` denotes the individual censoring time. In the presence of
        multiple transitions, :math:`\lambda(t)` is the sum over all possible
        transition-specific hazards:

        .. math::
            -\int_c^u \sum_{k'} \lambda^{k' \mid k}(t \mid t_0) \, dt,

        using the Chasles property to simplify computation and improve numerical
        precision.

        The input `u` must be a matrix of shape :math:`(n, m)` where :math:`n` is the
        number of individuals and :math:`m` is the number of prediction time points.

        Args:
            data (ModelData): The dataset containing covariates, observed outcomes,
                trajectories, and censoring information.
            u (torch.Tensor): Matrix of prediction times of shape `(n, m)`.
            n_samples (int, optional): Number of posterior samples to draw. Defaults to
                1000.
            double_monte_carlo (bool, optional): If True, predictions incorporate a
                double Monte Carlo procedure to sample parameters. Defaults to False.

        Raises:
            ValueError: If `double_monte_carlo` is True and the model has not been
                fitted.
            ValueError: If `u` contains NaN or infinite values.
            ValueError: If `u` has a shape inconsistent with the number of individuals.

        Returns:
            torch.Tensor: Predicted survival log-probabilities of shape
                `(n_samples, n, m)`, stacked along the first dimension.
        """
        assert_all_finite(u, input_name="u")
        u = torch.broadcast_to(u, (len(data), -1))

        # Load and complete data
        data = ModelDataUnchecked(
            data.x, data.t, data.y, data.trajectories, data.c
        ).prepare(self)

        # Initialize variables
        surv_logps_pred: list[torch.Tensor] = []
        n_iter = ceil(n_samples / self.n_chains)

        if double_monte_carlo:
            init_params = parameters_to_vector(self.params.parameters())
            sampled_params = self._sample_params(n_iter)

        # Initialize MCMC
        sampler = self._init_sampler(data).run(self.n_warmup)

        for i in trange(
            n_iter,
            desc="Predicting survival log probabilities",
            disable=not bool(self.verbose),
        ):
            if double_monte_carlo:
                vector_to_parameters(sampled_params[i], self.params.parameters())  # type: ignore

            indiv_params = self.design.indiv_params_fn(
                self.params.fixed_effects, data.x, sampler.b
            )
            sample_data = SampleData(data.x, data.trajectories, indiv_params, data.c)
            surv_logps = self.compute_surv_logps(sample_data, u)
            surv_logps_pred.extend(surv_logps[i] for i in range(surv_logps.size(0)))

            sampler.run(self.n_subsample)

        if double_monte_carlo:
            vector_to_parameters(init_params, self.params.parameters())  # type: ignore

        return torch.stack(surv_logps_pred[:n_samples])

    @torch.no_grad()  # type: ignore
    @validate_params(
        {
            "data": [ModelData],
            "c": [torch.Tensor],
            "max_length": [Interval(Integral, 1, None, closed="left")],
            "n_samples": [Interval(Integral, 1, None, closed="left")],
            "double_monte_carlo": [bool],
        },
        prefer_skip_nested_validation=True,
    )
    def predict_trajectories(
        self,
        data: ModelData,
        c: torch.Tensor,
        *,
        max_length: int = 10,
        n_samples: int = 1000,
        double_monte_carlo: bool = False,
    ) -> list[list[Trajectory]]:
        r"""Predict individual-level trajectories up to specified censoring times.

        Simulates the evolution of individual trajectories conditional on posterior
        draws of the random effects :math:`b`, up to the censoring times `c`.
        Trajectories are truncated to a maximum length of `max_length` to avoid
        infinite loops. The simulation algorithm is a variant of Gillespie's method
        adapted for individual parameters. If `double_monte_carlo` is True, then the
        prediction is computed using the double Monte Carlo procedure described in
        Rizopoulos (2011).

        The input `c` must be a column vector of shape :math:`(n, 1)` where :math:`n`
        is the number of individuals.

        Args:
            data (ModelData): The dataset containing covariates, observed outcomes,
                trajectories, and censoring information.
            c (torch.Tensor): Column vector of censoring times for each individual.
            max_length (int, optional): Maximum length of generated trajectories.
                Defaults to 10.
            n_samples (int, optional): Number of posterior samples to draw. Defaults to
                1000.
            double_monte_carlo (bool, optional): If True, predictions incorporate a
                double Monte Carlo procedure to sample parameters. Defaults to False.

        Raises:
            ValueError: If `double_monte_carlo` is True and the model has not been
                fitted.
            ValueError: If `c` contains NaN or infinite values.
            ValueError: If `c` has a shape inconsistent with the number of individuals.

        Returns:
            list[list[Trajectory]]: Predicted trajectories for each individual,
            organized as a list of lists, with the outer list indexing posterior draws
            and the inner list indexing individuals.
        """
        assert_all_finite(c, input_name="c")
        check_consistent_length(c, data)

        # Load and complete data
        data = ModelDataUnchecked(
            data.x, data.t, data.y, data.trajectories, data.c
        ).prepare(self)

        # Initialize variables
        trajectories_pred: list[list[Trajectory]] = []
        n_iter = ceil(n_samples / self.n_chains)

        if double_monte_carlo:
            init_params = parameters_to_vector(self.params.parameters())
            sampled_params = self._sample_params(n_iter)

        # Initialize MCMC
        sampler = self._init_sampler(data).run(self.n_warmup)

        for i in trange(
            n_iter,
            desc="Predicting trajectories",
            disable=not bool(self.verbose),
        ):
            if double_monte_carlo:
                vector_to_parameters(sampled_params[i], self.params.parameters())  # type: ignore

            # Sample trajectories, not possible to vectorize fully
            indiv_params = self.design.indiv_params_fn(
                self.params.fixed_effects, data.x, sampler.b
            )
            for j in range(indiv_params.size(0)):
                sample_data = SampleDataUnchecked(
                    data.x, data.trajectories, indiv_params[j], data.c
                )
                trajectories_pred.append(
                    self.sample_trajectories(sample_data, c, max_length=max_length)
                )

            sampler.run(self.n_subsample)

        if double_monte_carlo:
            vector_to_parameters(init_params, self.params.parameters())  # type: ignore

        return trajectories_pred[:n_samples]



from numbers import Integral
from typing import Any, Self, cast

import torch
from sklearn.base import BaseEstimator  # type: ignore
from sklearn.utils._param_validation import (  # type: ignore
    Interval,  # type: ignore
    StrOptions,  # type: ignore
    validate_params,  # type: ignore
)
from sklearn.utils.validation import assert_all_finite  # type: ignore
from torch import nn

from ..types._defs import LogBaseHazardFn
from ..utils._checks import check_matrix_dim
from ..utils._linalg import flat_from_log_cholesky, log_cholesky_from_flat


class PrecisionParameters(BaseEstimator, nn.Module):
    r"""`nn.Module` encapsulating precision matrix parameters.

    This class provides three types of precision matrix parametrization: full,
    diagonal, and spherical (scalar). The default is full matrix parametrization.
    Precision matrices are internally represented using the **log-Cholesky
    parametrization** of the inverse covariance (precision) matrix. Formally, let
    :math:`P = \Sigma^{-1}` be the precision matrix and :math:`L` its Cholesky factor
    with positive diagonal elements. The log-Cholesky representation :math:`\tilde{L}`
    is defined by:

    .. math::
        \tilde{L}_{ij} = L_{ij}, \quad i > j

    .. math::
        \tilde{L}_{ii} = \log L_{ii}.

    This representation ensures numerical stability and avoids explicit inversion when
    computing quadratic forms. The log determinant of the precision matrix is then

    .. math::
        \log \det P = 2 \operatorname{Tr}(\tilde{L}).

    Instances can be created from a precision matrix using the `from_precision` or from
    a covariance matrix using the `from_covariance` classmethod with `precision_type`
    set to `'full'`, `'diag'`, or `'spherical'`.

    Attributes:
        flat (torch.Tensor): Flat representation of the precision matrix suitable
            for optimization.
        dim (int): Dimension of the precision matrix.
        precision_type (str): Type of parametrization, one of `'full'`, `'diag'`,
            or `'spherical'`.

    Examples:
        >>> random_prec = PrecisionParameters.from_covariance(torch.eye(3), "diag")
        >>> noise_prec = PrecisionParameters.from_covariance(torch.eye(2), "spherical")
    """

    @classmethod
    @validate_params(
        {
            "P": [torch.Tensor],
            "precision_type": [StrOptions({"full", "diag", "spherical"})],
        },
        prefer_skip_nested_validation=True,
    )
    def from_precision(cls, P: torch.Tensor, precision_type: str = "full") -> Self:
        r"""Gets instance from precision matrix according to choice of precision type.

        Args:
            P (torch.Tensor): The square precision matrix.
            precision_type (str, optional): The method, `'full'`, `'diag'`, or
                `'spherical'`. Defaults to `'full'`.

        Returns:
            Self: The usable representation.
        """
        L = cast(torch.Tensor, torch.linalg.cholesky(P))  # type: ignore
        L.diagonal().log_()
        return cls(flat_from_log_cholesky(L, precision_type), L.size(0), precision_type)

    @classmethod
    @validate_params(
        {
            "V": [torch.Tensor],
            "precision_type": [StrOptions({"full", "diag", "spherical"})],
        },
        prefer_skip_nested_validation=True,
    )
    def from_covariance(cls, V: torch.Tensor, precision_type: str = "full") -> Self:
        r"""Gets instance from covariance matrix according to choice of precision type.

        Args:
            V (torch.Tensor): The square covariance matrix.
            precision_type (str, optional): The method, `'full'`, `'diag'`, or
                `'spherical'`. Defaults to `'full'`.

        Returns:
            Self: The usable representation.
        """
        return cls.from_precision(V.inverse(), precision_type)

    @validate_params(
        {
            "flat": [torch.Tensor],
            "dim": [Interval(Integral, 1, None, closed="left")],
            "precision_type": [StrOptions({"full", "diag", "spherical"})],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(self, flat: torch.Tensor, dim: int, precision_type: str):
        """Initializes the `PrecisionParameters` object.

        Args:
            flat (torch.Tensor): The flat representation of the precision matrix.
            dim (int): The dimension of the precision matrix.
            precision_type (str): The method used to parametrize the precision matrix.

        Raises:
            ValueError: If the representation is invalid.
        """
        super().__init__()  # type: ignore

        check_matrix_dim(flat, dim, precision_type)

        self.flat = nn.Parameter(flat)
        self.dim = dim
        self.precision_type = precision_type

    @property
    def precision(self) -> torch.Tensor:
        """Gets the precision matrix.

        Returns:
            torch.Tensor: The precision matrix.
        """
        L = log_cholesky_from_flat(self.flat, self.dim, self.precision_type)
        L.diagonal().exp_()
        return L @ L.T

    @property
    def covariance(self) -> torch.Tensor:
        """Gets the covariance matrix.

        Returns:
            torch.Tensor: The covariance matrix.
        """
        L = log_cholesky_from_flat(self.flat, self.dim, self.precision_type)
        L.diagonal().exp_()
        return torch.cholesky_inverse(L)

    @property
    def _prec_cholesky_and_log_eigvals(self) -> tuple[torch.Tensor, torch.Tensor]:
        """Gets Cholesky factor of precision matrix and its log eigvals.

        Returns:
            tuple[torch.Tensor, torch.Tensor]: Precision matrix and log eigvals.
        """
        L = log_cholesky_from_flat(self.flat, self.dim, self.precision_type)
        log_eigvals = 2 * L.diag()
        L.diagonal().exp_()

        return L, log_eigvals


class ModelParameters(BaseEstimator, nn.Module):
    r"""`nn.Module` encapsulating all model parameters for a multistate joint model.

    This module contains fixed population-level parameters, covariate effects, link
    coefficients, random effects, noise precision, and log base hazard functions.
    Parameters can be shared by assigning the same `nn.Parameter` object to multiple
    fields. Reusing tensors directly is not supported and requires wrapping in
    `nn.Parameter` for correct computations, as `nn.ParameterDict` would break the
    object ids otherwise.

    Fixed population-level parameters:
        - `fixed_effects` are fixed population-level parameters.

    Random effects and noise precision matrices:
        - `random_prec` and `noise_prec` are `PrecisionParameters` objects
          representing the random effects and residual noise precision matrices
          respectively.

    Base hazard functions:
        - `base_hazards` is a dictionary of `LogBaseHazardFn` modules keyed by
          `(from_state, to_state)` tuples for each transition; optimization can be
          disabled per hazard via its `frozen` attribute.

    Link and covariate coefficients:
        - `link_coefs` and `x_coefs` are dictionaries keyed by `(from_state,
          to_state)` tuples, representing linear coefficients for links and
          covariates, respectively.

    Attributes:
        fixed_effects (torch.Tensor): Fixed population-level parameters.
        random_prec (PrecisionParameters): Precision parameters for random effects.
        noise_prec (PrecisionParameters): Precision parameters for residual noise.
        base_hazards (nn.ModuleDict): Log base hazard functions per transition.
        link_coefs (nn.ParameterDict): Linear parameters for the link functions.
        x_coefs (nn.ParameterDict): Covariate parameters for each transition.

    Examples:
        >>> fixed_effects = torch.zeros(3)
        >>> random_prec = PrecisionParameters.from_covariance(torch.eye(3), "diag")
        >>> noise_prec = PrecisionParameters.from_covariance(torch.eye(2), "spherical")
        >>> link_coefs = {(0, 1): torch.zeros(3), (1, 0): torch.zeros(3)}
        >>> x_coefs = {(0, 1): torch.zeros(2), (1, 0): torch.zeros(2)}
        >>> params = ModelParameters(
        ...     fixed_effects,
        ...     random_prec,
        ...     noise_prec,
        ...     link_coefs,
        ...     x_coefs,
        ... )
        >>> # Shared parameters
        >>> shared_coef = nn.Parameter(torch.zeros(3))  # Mandatory nn.Parameter
        >>> shared_link_coefs = {(0, 1): shared_coef, (1, 0): shared_coef}
        >>> shared_params = ModelParameters(
        ...     fixed_effects,
        ...     random_prec,
        ...     noise_prec,
        ...     shared_link_coefs,
        ...     x_coefs,
        ... )
    """

    fixed_effects: torch.Tensor
    random_prec: PrecisionParameters
    noise_prec: PrecisionParameters
    base_hazards: nn.ModuleDict
    link_coefs: nn.ParameterDict
    x_coefs: nn.ParameterDict

    @validate_params(
        {
            "fixed_effects": [torch.Tensor],
            "random_prec": [PrecisionParameters],
            "noise_prec": [PrecisionParameters],
            "base_hazards": [dict],
            "link_coefs": [dict],
            "x_coefs": [dict],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(
        self,
        fixed_effects: torch.Tensor,
        random_prec: PrecisionParameters,
        noise_prec: PrecisionParameters,
        base_hazards: dict[tuple[Any, Any], LogBaseHazardFn],
        link_coefs: dict[tuple[Any, Any], torch.Tensor],
        x_coefs: dict[tuple[Any, Any], torch.Tensor],
    ):
        """Initializes the `ModelParams` object.

        Args:
            fixed_effects (torch.Tensor): Fixed population-level parameters.
            random_prec (PrecisionParameters): Precision parameters for random effects.
            noise_prec (PrecisionParameters): Precision parameters for residual noise.
            base_hazards (dict[tuple[Any, Any], LogBaseHazardFn]): Log base hazard
                functions.
            link_coefs (dict[tuple[Any, Any], torch.Tensor]): Linear parameters for the
                link functions.
            x_coefs (dict[tuple[Any, Any], torch.Tensor]): Covariate parameters for each
                transition.

        Raises:
            ValueError: If any of the tensors contains NaN or infinite values.
        """
        super().__init__()  # type: ignore

        self.fixed_effects = nn.Parameter(fixed_effects)
        self.random_prec = random_prec
        self.noise_prec = noise_prec
        self.base_hazards = nn.ModuleDict({str(k): v for k, v in base_hazards.items()})
        self.link_coefs = nn.ParameterDict({str(k): v for k, v in link_coefs.items()})
        self.x_coefs = nn.ParameterDict({str(k): v for k, v in x_coefs.items()})

        for key, val in self.named_parameters():
            assert_all_finite(val.detach(), input_name=key)

    def numel(self) -> int:
        """Return the number of unique parameters.

        Returns:
            int: The number of the (unique) parameters.
        """
        return sum(p.numel() for p in self.parameters())



from numbers import Integral
from typing import Any, cast

import numpy as np
import torch
from sklearn.utils._param_validation import Interval, validate_params  # type: ignore
from sklearn.utils.validation import (  #  type: ignore
    assert_all_finite,  #  type: ignore
    check_consistent_length,  #  type: ignore
)

from ..types._data import (
    ModelDataUnchecked,
    ModelDesign,
    SampleData,
    SampleDataUnchecked,
)
from ..types._defs import Trajectory
from ..types._parameters import ModelParameters
from ..utils._surv import build_remaining_buckets


class HazardMixin:
    """Mixin class for hazard model computations."""

    design: ModelDesign
    params: ModelParameters
    n_quad: int
    n_bisect: int

    def __init__(
        self,
        n_quad: int,
        n_bisect: int,
        *args: Any,
        **kwargs: Any,
    ):
        """Initializes the hazard mixin.

        Args:
            n_quad (int): Number of quadrature nodes.
            n_bisect (int): The number of bisection steps.
        """
        super().__init__(*args, **kwargs)

        self.n_quad = n_quad
        self.n_bisect = n_bisect
        self._std_nodes, self._std_weights = self._legendre_quad(n_quad)

    @staticmethod
    def _legendre_quad(n_quad: int) -> tuple[torch.Tensor, torch.Tensor]:
        """Get the Legendre quadrature nodes and weights.

        Args:
            n_quad (int): The number of quadrature points.

        Returns:
            tuple[torch.Tensor, torch.Tensor]: The nodes and weights.
        """
        nodes, weights = cast(
            tuple[
                np.ndarray[Any, np.dtype[np.float64]],
                np.ndarray[Any, np.dtype[np.float64]],
            ],
            np.polynomial.legendre.leggauss(n_quad),  # type: ignore
        )

        dtype = torch.get_default_dtype()
        std_nodes = torch.tensor(nodes, dtype=dtype).unsqueeze(0)
        std_weights = torch.tensor(weights, dtype=dtype)

        return std_nodes, std_weights

    def _log_hazard(
        self,
        key: tuple[Any, Any],
        t0: torch.Tensor,
        t1: torch.Tensor,
        x: torch.Tensor,
        indiv_params: torch.Tensor,
    ) -> torch.Tensor:
        """Computes log hazard.

        Args:
            key (tuple[Any, Any]): The transition key.
            t0 (torch.Tensor): A column vector of previous transition times.
            t1 (torch.Tensor): A matrix of next transition times.
            x (torch.Tensor): The fixed covariates.
            indiv_params (torch.Tensor): The individual parameters.

        Returns:
            torch.Tensor: The computed log hazard.
        """
        str_key = str(key)

        # Compute baseline hazard
        base = self.params.base_hazards[str_key](t0, t1)

        # Compute time-varying effects
        mod = (
            self.design.link_fns[key](t1, indiv_params)
            @ self.params.link_coefs[str_key]
        )

        # Compute covariates effect (if any)
        var = 0
        if str_key in self.params.x_coefs:
            var = x @ self.params.x_coefs[str_key].unsqueeze(-1)

        return (base + mod + var).reshape((*indiv_params.shape[:-1], -1))

    def _cum_hazard(
        self,
        key: tuple[Any, Any],
        t0: torch.Tensor,
        t1: torch.Tensor,
        x: torch.Tensor,
        indiv_params: torch.Tensor,
    ) -> torch.Tensor:
        """Computes cumulative hazard.

        Args:
            key (tuple[Any, Any]): The transition key.
            t0 (torch.Tensor): A column vector of previous transition times.
            t1 (torch.Tensor): A matrix of next transition times.
            x (torch.Tensor): The fixed covariates.
            indiv_params (torch.Tensor): The individual parameters.

        Returns:
            torch.Tensor: The computed cumulative hazard.
        """
        # Careful with negative times
        t1 = torch.max(t0, t1)

        # Transform to quadrature interval
        half = 0.5 * (t1 - t0)
        quad = (
            0.5 * (t0 + t1).unsqueeze(-1) + half.unsqueeze(-1) * self._std_nodes
        ).flatten(start_dim=-2)

        # Compute hazard at quadrature points
        vals = self._log_hazard(key, t0, quad, x, indiv_params).clamp(max=50).exp()

        return half * (
            vals.unflatten(-1, (-1, self._std_weights.size(-1))) @ self._std_weights  # type: ignore
        )

    def _hazard_logliks(
        self, data: ModelDataUnchecked, indiv_params: torch.Tensor
    ) -> torch.Tensor:
        """Computes the hazard log likelihoods.

        Args:
            data (ModelDataUnchecked): Dataset on which likelihood is computed.
            indiv_params (torch.Tensor): A matrix of individual parameters.

        Returns:
            torch.Tensor: The computed log likelihoods.
        """
        logliks = torch.zeros(indiv_params.shape[:-1])

        for key, (idxs, t0, obs, half, quad) in data.quad_buckets.items():
            vals = self._log_hazard(
                key, t0, quad, data.x[idxs], indiv_params[..., idxs, :]
            )
            vals[..., 1:].clamp_(max=50).exp_()

            # Compute log likelihoods and scatter add
            obs_logliks = vals[..., 0]
            alts_logliks = half.flatten() * (vals[..., 1:] @ self._std_weights)
            logliks.index_add_(-1, idxs, (-alts_logliks).addcmul(obs, obs_logliks))

        return logliks

    @torch.no_grad()  # type: ignore
    @validate_params(
        {
            "sample_data": [SampleData],
            "u": [torch.Tensor],
        },
        prefer_skip_nested_validation=True,
    )
    def compute_surv_logps(
        self, sample_data: SampleData, u: torch.Tensor
    ) -> torch.Tensor:
        r"""Compute log survival probabilities at specified times.

        Evaluates the log-probability of remaining event-free up to the prediction times
        :math:`u` conditional on individual-level parameters and censoring times. The
        computation uses the hazard function :math:`\lambda(t)` to obtain:

        .. math::
            \log \mathbb{P}(T^* \geq u \mid T^* > c) = -\int_c^u \lambda(t) \, dt,

        where :math:`c` denotes the censoring time for each individual. In cases with
        multiple possible transitions, :math:`\lambda(t)` sums over all
        transition-specific hazards:

        .. math::
            -\int_c^u \sum_{k'} \lambda^{k' \mid k}(t \mid t_0) \, dt,

        exploiting the Chasles property to simplify computation and improve numerical
        precision.

        The input `u` must be a matrix of shape :math:`(n, m)` where :math:`n` is the
        number of individuals and :math:`m` is the number of prediction time points.

        Args:
            sample_data (SampleData): The dataset containing covariates,
                individual-level parameters, trajectories, and censoring information.
            u (torch.Tensor): Matrix of evaluation times of shape `(n, m)`.

        Raises:
            ValueError: If `u` contains NaN or infinite values.
            ValueError: If `u` has a shape inconsistent with the number of individuals.

        Returns:
            torch.Tensor: Computed survival log-probabilities of shape `(n, m)`,
            with rows corresponding to individuals and columns to prediction times.
        """
        assert_all_finite(u, input_name="u")
        u = torch.broadcast_to(u, (len(sample_data), -1))

        x = sample_data.x
        indiv_params = sample_data.indiv_params
        t_cond = sample_data.t_cond

        # Get buckets from last states
        buckets = build_remaining_buckets(
            self, sample_data.trajectories, u.max(dim=1).values
        )

        # Compute the log probabilities summing over transitions
        nlogps = torch.zeros(*indiv_params.shape[:-1], u.size(1))
        for key, (idxs, t0) in buckets.items():
            # Compute negative log survival and scatter add
            t0 = t0 if t_cond is None else t_cond[idxs]  # noqa: PLW2901
            alts_logliks = self._cum_hazard(
                key, t0, u[idxs], x[idxs], indiv_params[..., idxs, :]
            )
            nlogps.index_add_(-2, idxs, alts_logliks)

        return -nlogps.clamp(min=0.0)

    def _sample_transition(
        self,
        key: tuple[Any, Any],
        t0: torch.Tensor,
        t1: torch.Tensor,
        x: torch.Tensor,
        indiv_params: torch.Tensor,
    ) -> torch.Tensor:
        """Sample survival times using inverse transform sampling.

        Args:
            key (tuple[Any, Any]): The transition key.
            t0 (torch.Tensor): The start times.
            t1 (torch.Tensor): The end times.
            x (torch.Tensor): The covariates.
            indiv_params (torch.Tensor): The individual parameters.

        Returns:
            torch.Tensor: The computed pre transition times.
        """
        # Initialize for bisection search
        t_left, t_right = (
            t0.clone(),
            torch.nextafter(t1, torch.tensor(torch.inf)),
        )

        # Generate exponential random variables
        target = -torch.log(torch.rand_like(t_left))

        # Bisection search for survival times
        for _ in range(self.n_bisect):
            t_mid = 0.5 * (t_left + t_right)
            surv_nlogps = self._cum_hazard(key, t0, t_mid, x, indiv_params)

            # Update search bounds
            accept_mask = surv_nlogps.reshape(target.shape) < target
            torch.where(accept_mask, t_mid, t_left, out=t_left)
            torch.where(accept_mask, t_right, t_mid, out=t_right)

        return t_right

    def _sample_trajectory_step(self, sample_data: SampleData, c: torch.Tensor) -> bool:
        """Appends the next simulated transition.

        Args:
            sample_data (SampleData): Sampling data
            c (torch.Tensor): Sampling censoring time.

        Returns:
            bool: True if the sampling is done.
        """
        x = sample_data.x
        indiv_params = sample_data.indiv_params
        t_cond = sample_data.t_cond

        # Get buckets from last states
        current_buckets = build_remaining_buckets(self, sample_data.trajectories, c)

        if not current_buckets:
            return True

        # Initialize candidate transition times
        n_transitions = len(current_buckets)
        t_candidates = torch.full((len(sample_data), n_transitions), torch.inf)

        for j, (key, (idxs, t0, t1)) in enumerate(current_buckets.items()):
            # Sample transition times, and condition with c
            t0 = t0 if t_cond is None else t_cond[idxs]  # noqa: PLW2901
            t_sample = self._sample_transition(
                key, t0, t1, x[idxs], indiv_params[..., idxs, :]
            )
            t_candidates[idxs, j] = t_sample.flatten()

        # Find earliest transition
        min_times, argmin_idxs = torch.min(t_candidates, dim=1)
        bucket_keys = list(current_buckets.keys())

        for i, (time, arg_idx) in enumerate(zip(min_times, argmin_idxs, strict=True)):
            if torch.isfinite(time):
                sample_data.trajectories[i].append(
                    (time.item(), bucket_keys[int(arg_idx)][1])
                )

        return False

    @torch.no_grad()  # type: ignore
    @validate_params(
        {
            "sample_data": [SampleData],
            "c": [torch.Tensor],
            "max_length": [Interval(Integral, 1, None, closed="left")],
        },
        prefer_skip_nested_validation=True,
    )
    def sample_trajectories(
        self,
        sample_data: SampleData,
        c: torch.Tensor,
        *,
        max_length: int = 10,
    ) -> list[Trajectory]:
        r"""Simulate individual trajectories from the multistate joint model.

        Generates sample trajectories for each individual up to the censoring times `c`,
        truncating to a maximum of `max_length` transitions to prevent infinite loops.
        The simulation employs a variant of Gillespie's algorithm adapted for individual
        parameter draws. These sampled trajectories form the basis for posterior
        predictive checks or downstream predictions in the joint model framework.

        The input `c` must be a column vector of shape :math:`(n, 1)` where :math:`n`
        is the number of individuals.

        Args:
            sample_data (SampleData): The dataset containing covariates, trajectories,
                and individual-level parameter used for sampling.
            c (torch.Tensor): Column vector of censoring times for each individual.
            max_length (int, optional): Maximum number of iterations or transitions
                sampled per trajectory. Defaults to 10.

        Raises:
            ValueError: If `c` contains NaN or infinite values.
            ValueError: If `c` has a shape inconsistent with the number of individuals.

        Returns:
            list[Trajectory]: List of sampled trajectories, one per individual, with
                each trajectory truncated at the censoring time.
        """
        assert_all_finite(c, input_name="c")
        check_consistent_length(c, sample_data)

        # Copy sample data to avoid modifying the original
        trajectories_copied = [
            trajectory.copy() for trajectory in sample_data.trajectories
        ]
        sample_data_copied = SampleDataUnchecked(
            sample_data.x,
            trajectories_copied,
            sample_data.indiv_params,
            sample_data.t_cond,
        )

        # Sample future transitions iteratively
        for _ in range(max_length):
            if self._sample_trajectory_step(sample_data_copied, c):
                break
            sample_data_copied.t_cond = None

        return [
            trajectory if trajectory[-1][0] <= c[i] else trajectory[:-1]
            for i, trajectory in enumerate(sample_data_copied.trajectories)
        ]



from typing import Any

import torch

from ..types._data import ModelDataUnchecked, ModelDesign
from ..types._defs import LOG_TWO_PI
from ..types._parameters import ModelParameters


class LongitudinalMixin:
    """Mixin class for longitudinal model computations."""

    design: ModelDesign
    params: ModelParameters

    def __init__(self, *args: Any, **kwargs: Any):
        """Initializes the longitudinal mixin."""
        super().__init__(*args, **kwargs)

    def _longitudinal_logliks(
        self, data: ModelDataUnchecked, indiv_params: torch.Tensor
    ) -> torch.Tensor:
        """Computes the longitudinal log likelihoods.

        Args:
            data (ModelDataUnchecked): Dataset on which likelihood is computed.
            indiv_params (torch.Tensor): A 3D tensor of individual parameters.

        Returns:
            torch.Tensor: The computed log likelihoods.
        """
        # Careful with NaNs
        predicted = self.design.regression_fn(data.valid_t, indiv_params)
        diff = data.valid_y.addcmul(predicted, data.valid_mask, value=-1.0)

        noise_prec_cholesky, noise_prec_log_eigvals = (
            self.params.noise_prec._prec_cholesky_and_log_eigvals  # type: ignore
        )
        noise_quad_form = (diff @ noise_prec_cholesky).pow(2).sum(dim=(-2, -1))
        noise_norm_factor = data.n_valid @ (noise_prec_log_eigvals - LOG_TWO_PI)

        return 0.5 * (noise_norm_factor - noise_quad_form)



from __future__ import annotations

from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from numbers import Integral
from typing import TYPE_CHECKING, Any, Self

import torch
from sklearn.base import BaseEstimator  # type: ignore
from sklearn.utils._param_validation import validate_params  # type: ignore
from sklearn.utils.validation import (  # type: ignore
    assert_all_finite,  # type: ignore
    check_consistent_length,  # type: ignore
)

from ..utils._checks import check_trajectories
from ..utils._surv import build_quad_buckets
from ._defs import IndividualParametersFn, LinkFn, RegressionFn, Trajectory

if TYPE_CHECKING:
    from ..model._fit import FitMixin
    from ..model._predict import PredictMixin


# Dataclasses
@dataclass
class ModelDesign(BaseEstimator):
    r"""Dataclass encapsulating the design of a multistate joint model.

    This class defines the parametric and functional design of the model, including
    individual-specific parameters, regression structures, and link functions for state
    transitions. All functions should be implemented to allow maximum broadcasting to
    ensure efficient vectorized computation. If broadcasting is not possible, `vmap` may
    be used for safe parallelization.

    Functions provided to the MCMC sampler will automatically be wrapped with
    `torch.no_grad()`. If gradient computation is required regardless of the sampling
    context, wrap the function explicitly with `torch.enable_grad()`.

    All functions must be well-defined on closed intervals of their input domain and
    differentiable almost everywhere to ensure compatibility with gradient-based
    procedures.

    Individual Parameters:
        - `indiv_params_fn` is a function that computes individual parameters. Given
          `fixed_effects` (fixed population-level parameters), `x` (covariates matrix of
          shape :math:`(n, p)`), and `b` (random effects, either 2D or 3D), it yields
          tensors of corresponding dimensions, either 2D or 3D depending on the model
          design. This function defines the mapping from population-level parameters and
          covariates to individual-specific parameters.

    Regression:
        - `regression_fn` is a function that maps time points and individual parameters
          to the expected observations. It must accept 1D or 2D time inputs and 2D or 3D
          individual parameters. The output tensor must have at least three dimensions:
          the last dimension corresponds to the response variable, the second-last to
          repeated measurements, the third-last to individuals, and an optional
          fourth-last dimension for parallelization across MCMC chains.

    Link Functions:
        - `link_fns` is a mapping from transition keys (tuples of
          `(state_from, state_to)`) to link functions. Each link function shares the
          same requirements as `regression_fn` and defines the transformation from
          regression outputs to transition-specific parameters.

    Attributes:
        indiv_params_fn (IndividualParametersFn): Function that computes individual
            parameters. Given `fixed_effects` (fixed population-level parameters), `x`
            (covariates matrix of shape :math:`(n, p)`), and `b` (random effects, either
            2D or 3D), it yields tensors of corresponding dimensions, either 2D or 3D
            depending on the model design. This function defines the mapping from
            population-level parameters and covariates to individual-specific
            parameters.
        regression_fn (RegressionFn): Regression function mapping time points and
            individual parameters to the expected observations. It must accept 1D or 2D
            time inputs and 2D or 3D individual parameters. The output tensor must have
            at least three dimensions: the last dimension corresponds to the response
            variable, the second-last to repeated measurements, the third-last to
            individuals, and an optional fourth-last dimension for parallelization
            across MCMC chains.
        link_fns (Mapping[tuple[Any, Any], LinkFn]): Mapping from transition keys
            (tuples of `(state_from, state_to)`) to link functions. Each link function
            shares the same requirements as `regression_fn` and defines the
            transformation from regression outputs to transition-specific parameters.

    Examples:
        >>> def sigmoid(t: torch.Tensor, indiv_params: torch.Tensor):
        ...     scale, offset, slope = indiv_params.chunk(3, dim=-1)
        ...     # Fully broadcasted
        ...     return (scale * torch.sigmoid((t - offset) / slope)).unsqueeze(-1)
        >>> fixed_plus_b = lambda fixed, x, b: fixed + b
        >>> link_fns = {("alive", "dead"): sigmoid}
        >>> design = ModelDesign(fixed_plus_b, sigmoid, link_fns)
    """

    indiv_params_fn: IndividualParametersFn
    regression_fn: RegressionFn
    link_fns: Mapping[tuple[Any, Any], LinkFn]


@dataclass
class ModelData(BaseEstimator):
    r"""Dataclass containing learnable multistate joint model data.

    It can be used with scikit-learn's train-test splits or cross-validation utilities.

    Covariates:
        - `x` is a matrix of covariates of shape :math:`(n, p)`, where :math:`n` denotes
          the number of individuals and :math:`p` the number of covariates.

    Measurement Times:
        - `t` represents the measurement times. It can be either:
            - a 2D tensor of shape :math:`(n, m)` when each individual has
              individual-specific time points,
            - a 1D tensor of shape :math:`(m,)` when all individuals share the same
              measurement times.

    Padding with NaNs is optional. However, `t` must not contain NaN values at positions
    where `y` is observed (i.e., where `y` is not NaN).

    Observations:
        - `y` is expected to be a 3D tensor of shape :math:`(n, m, d)`, where :math:`n`
          is the number of individuals, :math:`m` is the maximum number of measurements
          per individual, and :math:`d` is the dimension of the observation space
          :math:`\mathbb{R}^d`. Padding is performed with NaNs.

    Trajectories:
        - `trajectories` contains the individual-level multistate trajectories. They
          correspond to a `list[list[tuple[float, Any]]]` where each inner list is a
          trajectory and each tuple is a `(time, state)` pair.

    Censoring Times:
        - `c` represents the right censoring times, and are expected to be a column
          vector of shape :math:`(n, 1)`. Each value must be greater than or equal to
          the corresponding maximum transition time for each individual.

    The data can be completed using the `prepare` method, which formats it for manual
    likelihood evaluation and MCMC procedures. This usage is intended for advanced users
    familiar with the codebase. The method is called automatically by the `fit` and
    `predict` routines and does not require explicit user intervention in standard
    workflows.

    Raises:
        ValueError: If some trajectory is empty.
        ValueError: If some trajectory is not sorted.
        ValueError: If some trajectory is not compatible with the censoring times.
        ValueError: If any of the inputs contain inf or NaN values except `y`.
        ValueError: If the size is not consistent between inputs.

    Attributes:
        x (torch.Tensor): Fixed covariate matrix of shape `(n, p)`, where `n` is the
            number of individuals and `p` the number of covariates.
        t (torch.Tensor): Measurement times. Either a 1D tensor of shape `(m,)` when
            times are shared across individuals, or a 2D tensor of shape `(n, m)`
            when individuals have distinct time grids. Padding with NaNs may be
            used when required.
        y (torch.Tensor): Longitudinal measurements of shape `(n, m, d)`, where `n` is
            the number of individuals, `m` the maximum number of measurements per
            individual, and `d` the observation dimension. Padding is performed
            with NaNs when necessary.
        trajectories (list[Trajectory]): List of individual trajectories. Each
            `Trajectory` consists of a sequence of `(time, state)` tuples.
        c (torch.Tensor): Censoring times provided as a column vector. Each value must
            be greater than or equal to the corresponding maximum trajectory time.
        valid_mask (torch.Tensor): Boolean mask indicating valid (non-padded)
            measurements.
        n_valid (torch.Tensor): Number of valid measurements per individual.
        valid_t (torch.Tensor): Filtered tensor containing only valid measurement
            times.
        valid_y (torch.Tensor): Filtered tensor containing only valid measurements.
        buckets (dict[tuple[Any, Any], tuple[torch.Tensor, ...]]): Grouped trajectory
            data structures used for likelihood computation.
    """

    x: torch.Tensor
    t: torch.Tensor
    y: torch.Tensor
    trajectories: list[Trajectory]
    c: torch.Tensor

    def __len__(self) -> int:
        """Gets the number of individuals.

        Returns:
            int: The number of individuals.
        """
        return len(self.trajectories)

    def __getitem__(self, idxs: Integral | Sequence[Integral] | torch.Tensor) -> Self:
        """Gets the data for a single individual or a subset of individuals.

        Args:
            idxs (Integral | Sequence[Integral] | torch.Tensor): The index or indices of
                the individual(s).

        Returns:
            Self: The data for the individual(s).
        """
        idxs = torch.as_tensor(idxs).flatten()

        x_selected = self.x[idxs]
        t_selected = self.t[idxs]
        y_selected = self.y[idxs]
        trajectories_selected = [self.trajectories[i] for i in idxs]
        c_selected = self.c[idxs]

        return self.__class__(
            x=x_selected,
            t=t_selected,
            y=y_selected,
            trajectories=trajectories_selected,
            c=c_selected,
        )

    def __post_init__(self):
        """Runs the post init conversions.

        Raises:
            ValueError: If some trajectory is empty.
            ValueError: If some trajectory is not sorted.
            ValueError: If some trajectory is not compatible with the censoring times.
            ValueError: If any of the inputs contain NaN or infinite values except `t`
                and `y` for which NaN values are allowed.
            ValueError: If the size is not consistent between inputs.
        """
        validate_params(
            {
                "x": [torch.Tensor],
                "t": [torch.Tensor],
                "y": [torch.Tensor],
                "trajectories": [list],
                "c": [torch.Tensor],
            },
            prefer_skip_nested_validation=True,
        )

        check_trajectories(self.trajectories, self.c)

        assert_all_finite(self.x, input_name="x")
        assert_all_finite(self.t, input_name="t", allow_nan=True)
        assert_all_finite(self.y, input_name="y", allow_nan=True)
        assert_all_finite(self.c, input_name="c")

        check_consistent_length(self.x, self.y, self.c, self.trajectories)
        torch.broadcast_to(self.t, self.y.shape[:-1])

        # Check NaNs between t and y
        if ((~self.y.isnan()).any(dim=-1) & self.t.isnan()).any():
            raise ValueError("NaN time values on non NaN y values are not allowed")


@dataclass
class ModelDataUnchecked(ModelData):
    """Unchecked model data class."""

    valid_mask: torch.Tensor = field(init=False)
    n_valid: torch.Tensor = field(init=False)
    valid_t: torch.Tensor = field(init=False)
    valid_y: torch.Tensor = field(init=False)
    buckets: dict[tuple[Any, Any], tuple[torch.Tensor, ...]] = field(init=False)

    def __post_init__(self):
        """Overrides to skip checks."""
        pass

    def prepare(self, model: FitMixin | PredictMixin) -> Self:
        """Sets the representation for likelihood computations according to model.

        Args:
            model (FitMixin | PredictMixi): The model instance.

        Returns:
            Self: The prepared (completed) data.
        """
        self.valid_mask = ~self.y.isnan()
        self.n_valid = self.valid_mask.sum(dim=-2).to(self.y.dtype)
        self.valid_t = self.t.nan_to_num(self.t.nanmean().item())
        self.valid_y = self.y.nan_to_num()
        self.quad_buckets = build_quad_buckets(model, self.trajectories, self.c)

        return self


@dataclass
class SampleData(BaseEstimator):
    r"""Dataclass containing individual-level data for sampling procedures.

    It can be used with scikit-learn's train-test splits or cross-validation utilities.

    Covariates:
        - `x` is a matrix of covariates of shape :math:`(n, p)`, where :math:`n` denotes
          the number of individuals and :math:`p` the number of covariates.

    Trajectories:
        - `trajectories` contains the individual-level multistate trajectories. They
          correspond to a `list[list[tuple[float, Any]]]` where each inner list is a
          trajectory and each tuple is a `(time, state)` pair.

    Individual Parameters:
        - `indiv_params` represents the individual-specific parameters. It is expected
          to have the same number of rows as there are trajectories. Use a 3D tensor
          only if you fully understand the codebase and mechanisms. Trajectory sampling
          may only be used with matrices.

    Conditioning Times:
        - `t_cond` corresponds to truncation or conditioning times for each individual.
          This attribute is optional and, if not provided, is set to the maximum
          observation time per individual. This represents a landmark time up to which
          we condition on having observed no transitions. All simulated event times for
          the individual will be greater than or equal to `t_cond`.

    The data class is used for simulation and for prediction of quantities related to
    survival functions or trajectories. Unlike `ModelData`, it assumes exact knowledge
    of the individual parameters.

    Raises:
        ValueError: If some trajectory is empty.
        ValueError: If some trajectory is not sorted.
        ValueError: If some trajectory is not compatible with the conditioning times.
        ValueError: If any of the inputs contain NaN or infinite values.
        ValueError: If the size is not consistent between inputs.

    Attributes:
        x (torch.Tensor): Fixed covariate matrix of shape `(n, p)`, where `n` is the
            number of individuals and `p` the number of covariates.
        trajectories (list[Trajectory]): List of individual trajectories. Each
            `Trajectory` consists of a sequence of `(time, state)` tuples.
        indiv_params (torch.Tensor): Individual parameters with the same number of
            rows as there are trajectories. Use a matrix by default.
        t_cond (torch.Tensor | None): Optional conditioning times per individual. If
            None, the maximum observation time is used.
    """

    x: torch.Tensor
    trajectories: list[Trajectory]
    indiv_params: torch.Tensor
    t_cond: torch.Tensor | None = None

    def __len__(self) -> int:
        """Gets the number of individuals.

        Returns:
            int: The number of individuals.
        """
        return len(self.trajectories)

    def __getitem__(self, idxs: Integral | Sequence[Integral] | torch.Tensor) -> Self:
        """Gets the data for a single individual or a subset of individuals.

        Args:
            idxs (int | slice | Sequence[int] | torch.Tensor): The index or indices of
                the individual(s).

        Returns:
            Self: The data for the individual(s).
        """
        idxs = torch.as_tensor(idxs).flatten()

        x_selected = self.x[idxs]
        trajectories_selected = [self.trajectories[i] for i in idxs]
        indiv_params_selected = self.indiv_params[..., idxs, :]
        t_cond_selected = self.t_cond[idxs] if self.t_cond is not None else None

        return self.__class__(
            x=x_selected,
            trajectories=trajectories_selected,
            indiv_params=indiv_params_selected,
            t_cond=t_cond_selected,
        )

    def __post_init__(self):
        """Runs the post init conversions and checks.

        Raises:
            ValueError: If some trajectory is empty.
            ValueError: If some trajectory is not sorted.
            ValueError: If some trajectory is not compatible with the conditioning
                times.
            ValueError: If any of the inputs contain inf or NaN values.
            ValueError: If the size is not consistent between inputs.
        """
        validate_params(
            {
                "x": [torch.Tensor],
                "trajectories": [list],
                "indiv_params": [torch.Tensor],
                "t_cond": [torch.Tensor, None],
            },
            prefer_skip_nested_validation=True,
        )

        check_trajectories(self.trajectories, self.t_cond)

        assert_all_finite(self.x, input_name="x")
        assert_all_finite(self.indiv_params, input_name="indiv_params")
        assert_all_finite(self.t_cond, input_name="t_cond")

        check_consistent_length(
            self.x,
            self.indiv_params.transpose(0, -2),
            self.t_cond,
            self.trajectories,
        )


@dataclass
class SampleDataUnchecked(SampleData):
    """Unchecked sample data class."""

    def __post_init__(self):
        """Overrides to skip checks."""
        pass



from abc import ABC, abstractmethod
from typing import Final, NamedTuple, Protocol, TypeAlias

import torch
from torch import nn

# Type Aliases
Trajectory: TypeAlias = list[tuple[float, str]]


# Constants
LOG_TWO_PI: Final[torch.Tensor] = torch.log(torch.tensor(2.0 * torch.pi))


# User definitions
class LogBaseHazardFn(ABC, nn.Module):
    r"""Abstract base class for log base hazard functions.

    This class represents a log-transformed baseline hazard function in a multistate
    model. The log base hazard is parameterized as a `torch.nn.Module`, allowing its
    parameters to be optimized during model fitting. For default base hazards, a
    `frozen` attribute can be set to prevent optimization of the module parameters.

    Tensor input conventions:

    - `t0`: a column vector of previous transition times of shape :math:`(n, 1)`.
    - `t1`: a matrix of future time points at which the log base hazard is evaluated,
        of shape :math:`(n, m)` matching the number of rows in `t0`.

    Notes:
        The outputs are in log scale and can be directly used in likelihood
        computations for multistate models.
    """

    @abstractmethod
    def forward(self, t0: torch.Tensor, t1: torch.Tensor) -> torch.Tensor: ...


class IndividualParametersFn(Protocol):
    r"""Protocol defining the individual parameters function.

    This function maps population-level parameters, covariates, and random effects
    to individual-specific parameters used in a multistate model.

    Tensor input conventions:

    - `fixed_effects`: fixed population-level parameters.
    - `x`: covariates matrix of shape :math:`(n, p)`.
    - `b`: random effects of shape either :math:`(n, q)` (2D) or :math:`(n, m, q)` (3D).

    Tensor output conventions:

    - Last dimension corresponds to the number of parameters :math:`l`.
    - Second-last dimension corresponds to the number of individuals :math:`n`.
    - Optional third-last dimension may be used for parallelization across MCMC
      chains (= batched processing).

    Args:
        fixed_effects (torch.Tensor): Fixed population-level parameters.
        x (torch.Tensor): Fixed covariates matrix.
        b (torch.Tensor): Random effects tensor.

    Returns:
        torch.Tensor: Individual parameters tensor of shape consistent with
            :math:`(n, l)` or :math:`(n_chains, n, l)` for parallelized
            computations.

    Examples:
        >>> indiv_params_fn = lambda fixed, x, b: fixed + b
    """

    def __call__(
        self, fixed_effects: torch.Tensor, x: torch.Tensor, b: torch.Tensor
    ) -> torch.Tensor: ...


class RegressionFn(Protocol):
    r"""Protocol defining a regression function for multistate models.

    This function maps evaluation times and individual-specific parameters to predicted
    response values. It must support both 1D and 2D time inputs and individual
    parameters of order 2 or 3, returning either 3D or 4D tensors depending on the
    model design.

    Tensor input conventions:

    - `t` represents the measurement times. It can be either:
        - a 2D tensor of shape :math:`(n, m)` when each individual has
          individual-specific time points,
        - a 1D tensor of shape :math:`(m,)` when all individuals share the same
          measurement times.
    - `indiv_params` represents the individual-specific parameters. It is expected
      to have shape :math:`(n, l)` or :math:`(n_chains, n, l)` where :math:`n`
      is the number of individuals and :math:`l` is the number of parameters.

    Tensor output conventions:

    - Last dimension corresponds to the response variable dimension :math:`d`.
    - Second-last dimension corresponds to repeated measurements :math:`m`.
    - Third-last dimension corresponds to individual index :math:`n`.
    - Optional fourth-last dimension may be used for parallelization across MCMC
      chains (= batched processing).

    This protocol is conceptually identical to `LinkFn`.

    Args:
        t (torch.Tensor): Evaluation times of shape :math:`(n, m)` or :math:`(m,)`.
        indiv_params (torch.Tensor): Individual parameters of shape 2D :math:`(n, l)`
            or 3D :math:`(n_chains, n, l)`.

    Returns:
        torch.Tensor: Predicted response values of shape consistent with `(n, m, d)` or
            `(n_chains, n, m, d)` for parallelized computations.

    Examples:
        >>> def sigmoid(t: torch.Tensor, indiv_params: torch.Tensor):
        ...     scale, offset, slope = indiv_params.chunk(3, dim=-1)
        ...     # Fully broadcasted computation
        ...     return (scale * torch.sigmoid((t - offset) / slope)).unsqueeze(-1)
    """

    def __call__(self, t: torch.Tensor, indiv_params: torch.Tensor) -> torch.Tensor: ...


class LinkFn(Protocol):
    r"""Protocol defining a link function for multistate models.

    A link function maps evaluation times and individual-specific parameters to
    transformed outputs, such as transition-specific parameters. Requirements are
    identical to those of `RegressionFn`.

    Tensor input conventions:

    - `t` represents the measurement times. It can be either:
        - a 2D tensor of shape :math:`(n, m)` when each individual has
          individual-specific time points,
        - a 1D tensor of shape :math:`(m,)` when all individuals share the same
          measurement times.
    - `indiv_params` represents the individual-specific parameters. It is expected
      to have shape :math:`(n, l)` or :math:`(n_chains, n, l)` where :math:`n`
      is the number of individuals and :math:`l` is the number of parameters.

    Tensor output conventions:

    - Last dimension corresponds to the response variable dimension :math:`d`.
    - Second-last dimension corresponds to repeated measurements :math:`m`.
    - Third-last dimension corresponds to individual index :math:`n`.
    - Optional fourth-last dimension may be used for parallelization across MCMC
      chains (= batched processing).

    This protocol is conceptually identical to `RegressionFn`.

    Args:
        t (torch.Tensor): Evaluation times of shape :math:`(n, m)` or :math:`(m,)`.
        indiv_params (torch.Tensor): Individual parameters of shape 2D :math:`(n, l)`
            or 3D :math:`(n_chains, n, l)`.

    Returns:
        torch.Tensor: Transformed outputs consistent with shapes `(n, m, d)` or
            `(n_chains, n, m, d)` for parallelized computations.

    Examples:
        >>> def sigmoid(t: torch.Tensor, indiv_params: torch.Tensor):
        ...     scale, offset, slope = indiv_params.chunk(3, dim=-1)
        ...     # Fully broadcasted computation
        ...     return (scale * torch.sigmoid((t - offset) / slope)).unsqueeze(-1)
    """

    def __call__(self, t: torch.Tensor, indiv_params: torch.Tensor) -> torch.Tensor: ...


# Named tuples
class BucketData(NamedTuple):
    r"""NamedTuple representing a set of transitions for visualization purposes.

    This structure stores the transition times of individuals grouped together,
    typically used to visualize the trajectories per transition type in multistate
    models. Each entry corresponds to a single transition for a specific individual.

    Attributes:
        idxs (torch.Tensor): Indices of individuals corresponding to the transitions in
            this bucket, shape :math:`(k,)`, where :math:`k` is the number of
            transitions in the bucket.
        t0 (torch.Tensor): Column vector of previous transition times, shape
            :math:`(k, 1)`. Represents the start time of each transition.
        t1 (torch.Tensor): Column vector of next transition times, shape
            :math:`(k, 1)`. Represents the end time of each transition.

    Notes:
        Each tensor is aligned such that `t0[i]` and `t1[i]` correspond to the
        transition of individual `idxs[i]`. This alignment facilitates plotting or
        analyzing transitions per type across individuals.
    """

    idxs: torch.Tensor
    t0: torch.Tensor
    t1: torch.Tensor



from math import ceil
from typing import Any, Self
from warnings import warn

import torch
from sklearn.base import check_is_fitted  # type: ignore
from sklearn.utils._param_validation import validate_params  # type: ignore
from torch import nn
from torch.func import jacfwd  # type: ignore
from torch.nn.utils import parameters_to_vector
from torch.nn.utils.stateless import _reparametrize_module  # type: ignore
from tqdm import trange

from ..types._data import ModelData, ModelDataUnchecked, ModelDesign
from ..types._parameters import ModelParameters
from ._hazard import HazardMixin
from ._longitudinal import LongitudinalMixin
from ._prior import PriorMixin
from ._sampler import MCMCMixin, MetropolisWithinGibbsSampler


class FitMixin(PriorMixin, LongitudinalMixin, HazardMixin, MCMCMixin, nn.Module):
    """Mixin for fitting the model."""

    design: ModelDesign
    params: ModelParameters
    optimizer: torch.optim.Optimizer | None
    sampler: MetropolisWithinGibbsSampler | None
    n_warmup: int
    n_subsample: int
    max_iter: int
    tol: float
    window_size: int
    verbose: bool | int
    params_history_: list[torch.Tensor]
    fim_: torch.Tensor | None
    loglik_: float | None
    aic_: float | None
    bic_: float | None

    def __init__(
        self,
        optimizer: torch.optim.Optimizer | None,
        max_iter: int,
        tol: float,
        window_size: int,
        *args: Any,
        **kwargs: Any,
    ):
        """Initializes the fit parameters.

        Args:
            optimizer (torch.optim.Optimizer): The optimizer.
            max_iter (int): The maximum number of iterations for fitting.
            tol (float): The tolerance for the convergence.
            window_size (int): The window size for the convergence.
        """
        super().__init__(*args, **kwargs)

        self.optimizer = optimizer
        self.sampler = None
        self.max_iter = max_iter
        self.tol = tol
        self.window_size = window_size

    def _logpdfs_fn(
        self,
        data: ModelDataUnchecked,
        b: torch.Tensor,
    ) -> torch.Tensor:
        """Gets the log pdfs.

        Args:
            data (ModelData): Dataset on which likelihood is computed.
            b (torch.Tensor): The random effects.

        Returns:
           torch.Tensor: The log pdfs.
        """
        indiv_params = self.design.indiv_params_fn(self.params.fixed_effects, data.x, b)
        return (
            self._longitudinal_logliks(data, indiv_params)
            + self._hazard_logliks(data, indiv_params)
            + self._prior_logliks(b)
        )

    def _is_converged(self) -> bool:
        """Checks if the optimizer has converged.

        This is based on a linear regression of the parameters with the current
        number of iterations. If the mean of :math:`R^2` is below a threshold,
        the optimizer is considered to have converged.

        Returns:
            bool: True if the optimizer has converged, False otherwise.
        """

        def r2(Y: torch.Tensor) -> torch.Tensor:
            n = Y.size(0)
            i = torch.arange(n, dtype=torch.get_default_dtype())
            i_centered = i - (n - 1) / 2
            y_centered = Y - Y.mean(dim=0)
            num = (i_centered @ y_centered) ** 2
            den = i_centered.pow(2).sum() * y_centered.pow(2).sum(dim=0)
            return (num / den).nan_to_num()

        if len(self.params_history_) < self.window_size:
            return False

        Y = torch.stack(self.params_history_[-self.window_size :])
        return r2(Y).mean().item() < self.tol

    def _fit(self, data: ModelDataUnchecked, sampler: MetropolisWithinGibbsSampler):
        """Fits the model using the optimizer and the sampler.

        Args:
            data (ModelData): The data.
            sampler (MetropolisWithinGibbsSampler): The sampler.

        Raises:
            ValueError: If the optimizer is not initialized.
        """
        if self.max_iter <= 0:
            return

        if self.optimizer is None:
            raise ValueError("Optimizer is not initialized.")

        def closure():
            self.optimizer.zero_grad()  # type: ignore
            loss = -self._logpdfs_fn(data, sampler.b).mean()
            loss.backward()  # type: ignore
            return loss.item()

        for i in trange(  # noqa: B007
            self.max_iter,
            desc="Fitting joint model",
            disable=not bool(self.verbose),
        ):
            self.optimizer.step(closure)
            self.params_history_.append(
                parameters_to_vector(self.params.parameters()).detach()
            )

            # Restore logpdfs and indiv_params, because parameters changed
            sampler.reset().run(self.n_subsample)

            if self._is_converged():
                break

        if i == self.max_iter - 1:  # type: ignore
            warn(
                "Model may not have converged in the specified number of iterations. "
                "Try to increase `max_iter`, `tol`, or `window_size`. Also try "
                "to increase `n_subsample` or `n_warmup` for better MCMC mixing.",
                stacklevel=2,
            )

    @validate_params(
        {
            "data": [ModelData],
        },
        prefer_skip_nested_validation=True,
    )
    def fit(self, data: ModelData) -> Self:
        r"""Fit the model to observed data using maximum likelihood estimation.

        Computes the Maximum Likelihood Estimate (MLE) :math:`\hat{\theta}` of the model
        parameters. Optimization is performed using the configured `optimizer` for up to
        `max_iter` iterations. Convergence is assessed via a linearity-based
        stationarity test on the last `window_size` iterates: the :math:`R^2` statistic
        measures whether the trajectory of each parameter component is better explained
        by a linear trend than by a constant. Convergence is declared when all
        :math:`R^2` values are below `tol`, indicating negligible linear drift.

        The fitting procedure leverages the Fisher identity coupled with a stochastic
        gradient algorithm and a Metropolis-Hastings MCMC sampler. The Fisher identity
        states:

        .. math::
            \nabla_\theta \log \mathcal{L}(\theta ; x) = \mathbb{E}_{b \sim p(\cdot
            \mid x, \theta)} \left( \nabla_\theta \log \mathcal{L}(\theta ; x, b)
            \right).

        Args:
            data (ModelData): Dataset containing covariates, longitudinal measurements,
                trajectories, and censoring times used for fitting.

        Raises:
            ValueError: If the optimizer has not been initialized prior to fitting.

        Returns:
            Self: The fitted model instance with estimated parameters.
        """
        data = ModelDataUnchecked(
            data.x, data.t, data.y, data.trajectories, data.c
        ).prepare(self)

        # Initialize MCMC
        self.sampler = self._init_sampler(data).run(self.n_warmup)

        if self.optimizer is None:
            raise ValueError("Optimizer is not initialized.")

        def closure():
            self.optimizer.zero_grad()  # type: ignore
            loss = -self.sampler.logpdfs_fn(self.sampler.b).mean()  # type: ignore
            loss.backward()  # type: ignore
            return loss.item()

        i = 0
        for i in trange(  # noqa: B007
            self.max_iter,
            desc="Fitting joint model",
            disable=not bool(self.verbose),
        ):
            self.optimizer.step(closure)
            self.params_history_.append(
                parameters_to_vector(self.params.parameters()).detach()
            )

            # Restore logpdfs and indiv_params, because parameters changed
            self.sampler.reset().run(self.n_subsample)

            if self._is_converged():
                break

        if i == self.max_iter - 1:
            warn(
                "Model may not have converged in the specified number of iterations. "
                "Try to increase `max_iter`, `tol`, or `window_size`. Also try "
                "to increase `n_subsample` or `n_warmup` for better MCMC mixing.",
                stacklevel=2,
            )

        return self

    def compute_summary(self, n_samples: int = 500) -> Self:
        r"""Computes summary statistics for the fitted model.

        The expected Fisher Information Matrix is estimated as:

        .. math::
            \mathcal{I}_n(\theta) = \sum_{i=1}^n \mathbb{E}_{b \sim p(\cdot \mid x_i,
            \hat{\theta})} \left(\nabla \log \mathcal{L}(\hat{\theta} ; x_i, b) \nabla
            \log \mathcal{L}(\hat{\theta} ; x_i, b)^T \right).

        Model selection criteria are computed using a Laplace approximation of the
        posterior distribution, providing closed-form estimates of entropy. The
        parameter `n_samples` controls the number of posterior samples used for
        computing the Fisher Information Matrix and selection metrics; increasing this
        number improves accuracy at the cost of computational time.

        For additional details, see ISSN 2824-7795.

        Args:
            n_samples (int, optional): Number of posterior samples to use for computing
                the Fisher Information Matrix and selection metrics. Defaults to 500.

        Returns:
            Self: The fitted model with summary statistics computed.
        """
        check_is_fitted(self, "sampler")

        n, q = self.sampler.b.shape[1:]  # type: ignore

        # Jac forward since output dimension > input dimension
        @jacfwd  # type: ignore
        def _dict_jac_fn(
            named_parameters_dict: dict[str, torch.Tensor],
        ) -> torch.Tensor:
            with _reparametrize_module(self, named_parameters_dict):
                return self.sampler.logpdfs_fn(self.sampler.b).mean(dim=0)  # type: ignore

        def _jac_fn() -> torch.Tensor:
            out = _dict_jac_fn(dict(self.named_parameters()))  # type: ignore
            return torch.cat([p.reshape(n, -1) for p in out.values()], dim=-1)  # type: ignore

        # Initialize accumulators
        mjac = torch.zeros(n, self.params.numel())
        logpdf = 0.0
        mb = torch.zeros(n, q)
        mb2 = torch.zeros(n, q, q)

        n_iter = ceil(n_samples / self.n_chains)
        for _ in trange(
            n_iter,
            desc="Computing FIM and Model Selection Criteria",
            disable=not bool(self.verbose),
        ):
            # Mean jacobian across chains
            mjac += _jac_fn().detach()  # type: ignore

            # Mean logpdf across chains
            logpdf += self.sampler.logpdfs.sum().item() / self.n_chains  # type: ignore

            # Mean and outer product of b across chains
            mb += self.sampler.b.mean(dim=0)  # type: ignore
            mb2 += (
                torch.einsum("ijk,ijl->jkl", self.sampler.b, self.sampler.b)  # type: ignore
                / self.n_chains
            )

            self.sampler.run(self.n_subsample)  # type: ignore

        mjac /= n_iter
        logpdf /= n_iter
        mb /= n_iter
        mb2 /= n_iter

        # Compute FIM as variance of the score
        self.fim_ = mjac.T @ mjac

        # Compute entropy Laplace approximation
        covs = mb2 - torch.einsum("ij,ik->ijk", mb, mb)
        entropy = 0.5 * (torch.logdet(covs) + self.params.random_prec.dim).sum().item()

        self.loglik_ = logpdf + entropy
        self.aic_ = -2 * self.loglik_ + 2 * self.params.numel()
        self.bic_ = -2 * self.loglik_ + torch.logdet(self.fim_).item()

        return self



from __future__ import annotations

import itertools
from array import array
from collections import defaultdict
from functools import lru_cache
from typing import TYPE_CHECKING, Any

import torch
from sklearn.utils._param_validation import validate_params  # type: ignore

from ..types._defs import BucketData, Trajectory

if TYPE_CHECKING:
    from ..model._hazard import HazardMixin


@validate_params(
    {"trajectories": [list]},
    prefer_skip_nested_validation=True,
)
def build_buckets(
    trajectories: list[Trajectory],
) -> dict[tuple[Any, Any], BucketData]:
    """Builds buckets from trajectories for user convenience.

    The return structure stores the transition times of individuals grouped together,
    typically used to visualize the trajectories per transition type in multistate
    models. Each entry corresponds to a single transition for a specific individual.

    Args:
        trajectories (list[Trajectory]): The list of individual trajectories.

    Raises:
        TypeError: If the default dtype is not `float32` or `float64`.

    Returns:
        dict[tuple[Any, Any], BucketData]: Transition keys with values `BucketData`.
    """
    dtype = torch.get_default_dtype()
    if dtype == torch.float32:
        typecode = "f"
    elif dtype == torch.float64:
        typecode = "d"
    else:
        raise TypeError(f"Unsupported default dtype: {dtype}")

    # Process each individual trajectory
    buckets: defaultdict[
        tuple[Any, Any], tuple[array[int], array[float], array[float]]
    ] = defaultdict(lambda: (array("q"), array(typecode), array(typecode)))

    for i, trajectory in enumerate(trajectories):
        for (t0, s0), (t1, s1) in itertools.pairwise(trajectory):
            idxs, t0s, t1s = buckets[(s0, s1)]
            idxs.append(i)
            t0s.append(t0)
            t1s.append(t1)

    result = {
        key: BucketData(
            torch.frombuffer(idxs, dtype=torch.int64),
            torch.frombuffer(t0s, dtype=dtype).reshape(-1, 1),
            torch.frombuffer(t1s, dtype=dtype).reshape(-1, 1),
        )
        for key, (idxs, t0s, t1s) in buckets.items()
    }

    return dict(sorted(result.items()))


@lru_cache
def _build_alt_map(
    surv_keys: tuple[tuple[Any, Any], ...],
) -> defaultdict[Any, tuple[tuple[Any, Any], ...]]:
    """Builds alternative state mapping as tuples in a defaultdict.

    Args:
        surv_keys (tuple[tuple[Any, Any], ...]): The survival keys.

    Returns:
        defaultdict[Any, tuple[tuple[Any, Any], ...]]: The alternative state map.
    """
    return defaultdict(
        lambda: (),
        {
            s0: tuple((k, v) for k, v in surv_keys if k == s0)
            for s0 in {s0 for s0, _ in surv_keys}
        },
    )


def build_quad_buckets(
    model: HazardMixin,
    trajectories: list[Trajectory],
    c: torch.Tensor,
) -> dict[tuple[Any, Any], tuple[torch.Tensor, ...]]:
    """Build vectorizable bucket representation.

    Args:
        model (HazardMixin): The model instance.
        trajectories (list[Trajectory]): The trajectories.
        c (torch.Tensor): Censoring times.

    Raises:
        TypeError: If the default dtype is not `float32` or `float64`.

    Returns:
        dict[tuple[Any, Any], tuple[torch.Tensor, ...]]: The vectorizable buckets
            representation.
    """
    alt_map = _build_alt_map(tuple(model.design.link_fns.keys()))
    dtype = torch.get_default_dtype()
    if dtype == torch.float32:
        typecode = "f"
    elif dtype == torch.float64:
        typecode = "d"
    else:
        raise TypeError(f"Unsupported default dtype: {dtype}")

    # Initialize buckets
    buckets: defaultdict[
        tuple[Any, Any], tuple[array[int], array[float], array[float], array[bool]]
    ] = defaultdict(lambda: (array("q"), array(typecode), array(typecode), array("b")))

    # Process each individual trajectory
    for i, trajectory in enumerate(trajectories):
        for (t0, s0), (t1, s1) in itertools.pairwise(trajectory):
            for key in alt_map[s0]:
                idxs, t0s, t1s, obs = buckets[key]
                idxs.append(i)
                t0s.append(t0)
                t1s.append(t1)
                obs.append(key[1] == s1)

        (last_t, last_s), c_i = trajectory[-1], c[i].item()

        if last_t >= c_i:
            continue

        for key in alt_map[last_s]:
            idxs, t0s, t1s, obs = buckets[key]
            idxs.append(i)
            t0s.append(last_t)
            t1s.append(c_i)
            obs.append(False)

    out: dict[tuple[Any, Any], tuple[torch.Tensor, ...]] = {}
    for key, (idxs, t0s, t1s, obs) in buckets.items():
        idxs_ = torch.frombuffer(idxs, dtype=torch.int64)
        t0_ = torch.frombuffer(t0s, dtype=dtype).reshape(-1, 1)
        t1_ = torch.frombuffer(t1s, dtype=dtype).reshape(-1, 1)
        obs_ = torch.frombuffer(obs, dtype=torch.bool)
        half = 0.5 * (t1_ - t0_)
        quad = torch.cat(
            [t1_, (t0_ + t1_).addmm(half, model._std_nodes, beta=0.5)],  # type: ignore
            dim=-1,
        )

        out[key] = (idxs_, t0_, obs_, half, quad)

    return out


def build_remaining_buckets(
    model: HazardMixin,
    trajectories: list[Trajectory],
    c: torch.Tensor,
) -> dict[tuple[Any, Any], tuple[torch.Tensor, ...]]:
    """Build possible bucket representation.

    Args:
        model (HazardMixin): The model instance.
        trajectories (list[Trajectory]): The trajectories.
        c (torch.Tensor): Censoring times.

    Raises:
        TypeError: If the default dtype is not `float32` or `float64`.

    Returns:
        dict[tuple[Any, Any], tuple[torch.Tensor, ...]]: The possible buckets
            representation.
    """
    alt_map = _build_alt_map(tuple(model.design.link_fns.keys()))
    dtype = torch.get_default_dtype()
    if dtype == torch.float32:
        typecode = "f"
    elif dtype == torch.float64:
        typecode = "d"
    else:
        raise TypeError(f"Unsupported default dtype: {dtype}")

    # Initialize buckets
    buckets: defaultdict[tuple[Any, Any], tuple[array[int], array[float]]] = (
        defaultdict(lambda: (array("q"), array(typecode)))
    )

    # Process each individual trajectory
    for i, trajectory in enumerate(trajectories):
        last_t, last_s = trajectory[-1]

        if last_t >= c[i].item():
            continue

        for key in alt_map[last_s]:
            idxs, t0s = buckets[key]
            idxs.append(i)
            t0s.append(last_t)

    return {
        key: (
            idxs_tensor := torch.frombuffer(idxs, dtype=torch.int64),
            torch.frombuffer(t0s, dtype=dtype).reshape(-1, 1),
            c[idxs_tensor],
        )
        for key, (idxs, t0s) in buckets.items()
    }
